import io
import secrets
import time
from typing import TYPE_CHECKING, Dict, Generic, Literal, Type, TypeVar, Union

import aiohttp
from lonelypsp.auth.config import AuthResult
from lonelypsp.compat import fast_dataclass
from lonelypsp.stateless.constants import BroadcasterToSubscriberStatelessMessageType
from lonelypsp.tracing.stateless.notify import (
    StatelessTracingNotifyOnResponseReceived,
    StatelessTracingNotifyOnRetryDetermined,
    StatelessTracingNotifyOnSending,
)

from lonelypsc.config.config import PubSubBroadcasterConfig
from lonelypsc.config.http_config import HttpPubSubConfig
from lonelypsc.http.notify.result import HttpPubSubNotifyResult
from lonelypsc.http.retrier import (
    BroadcasterRetryableAction,
    BroadcasterRetryableActionResult,
    BroadcasterRetryableActionResultFailure,
    BroadcasterRetryableActionResultRetry,
    BroadcasterRetryableActionResultSuccess,
    BroadcasterRetryableActionRetryFollowup,
)
from lonelypsc.util.async_io import async_read_exact
from lonelypsc.util.io_helpers import (
    PositionedSyncStandardIO,
    PrefixedSyncStandardIO,
    SyncStandardWithLengthIO,
)

InitializerT = TypeVar("InitializerT")


@fast_dataclass
class HttpNotifyInfo:
    """The information about the underlying notification"""

    topic: bytes
    """the topic the message is being posted to"""
    normalized_message: SyncStandardWithLengthIO
    """the message to send where offset 0 is the start and length is the end"""
    sha512: bytes
    """the sha512 of the message"""


class HttpNotifyAction(Generic[InitializerT]):
    """The action taken to post a new message to a topic"""

    def __init__(
        self,
        /,
        *,
        config: HttpPubSubConfig[InitializerT],
        info: HttpNotifyInfo,
        session: aiohttp.ClientSession,
        tracer: StatelessTracingNotifyOnSending,
        seen_ambiguous: bool,
    ):
        self.config = config
        self.info = info
        self.session = session
        self.tracer = tracer
        self.seen_ambiguous = seen_ambiguous

    async def attempt(self, /, *, broadcaster: PubSubBroadcasterConfig) -> Union[
        BroadcasterRetryableActionResultSuccess[HttpPubSubNotifyResult],
        BroadcasterRetryableActionResultFailure[Literal["ambiguous", "refused"]],
        BroadcasterRetryableActionResultRetry[
            HttpPubSubNotifyResult, Literal["ambiguous", "refused"]
        ],
    ]:
        identifier = secrets.token_bytes(4)
        tracing_and_followup = self.tracer.on_sending_request(
            broadcaster=broadcaster["host"], identifier=identifier
        )
        tracing = tracing_and_followup.tracing
        on_response_received_tracer = tracing_and_followup.followup

        auth_at = time.time()
        authorization = await self.config.authorize_notify(
            tracing=tracing,
            topic=self.info.topic,
            identifier=identifier,
            message_sha512=self.info.sha512,
            now=auth_at,
        )

        message_prefix = io.BytesIO()
        message_prefix.write(len(self.info.topic).to_bytes(2, "big", signed=False))
        message_prefix.write(self.info.topic)
        message_prefix.write(self.info.sha512)
        message_prefix.write(len(tracing).to_bytes(2, "big", signed=False))
        message_prefix.write(tracing)
        message_prefix.write(len(identifier).to_bytes(1, "big", signed=False))
        message_prefix.write(identifier)
        message_prefix.write(
            len(self.info.normalized_message).to_bytes(8, "big", signed=False)
        )

        body = PrefixedSyncStandardIO(
            PositionedSyncStandardIO(message_prefix, 0, message_prefix.tell()),
            self.info.normalized_message,
        )
        headers: Dict[str, str] = {
            "Content-Type": "application/octet-stream",
            "Content-Length": str(len(body)),
        }
        if authorization is not None:
            headers["Authorization"] = authorization

        try:
            result = await self.session.post(
                broadcaster["host"] + "/v1/notify",
                data=body,
                headers=headers,
                allow_redirects=False,
                read_until_eof=False,
            )
        except aiohttp.ClientError:
            return await self._on_network_error(
                on_response_received_tracer=on_response_received_tracer
            )

        try:
            await result.__aenter__()
        except BaseException as e:
            await result.__aexit__(type(e), e, None)
            await self._on_network_error(
                on_response_received_tracer=on_response_received_tracer
            )
            raise

        try:
            on_response_auth_result = on_response_received_tracer.on_response_received(
                status_code=result.status
            )
            del on_response_received_tracer
        except BaseException as e:
            await result.__aexit__(type(e), e, None)
            raise

        try:
            type_bytes = await async_read_exact(result.content, 2)
            type_int = int.from_bytes(type_bytes, "big", signed=False)
            if type_int != int(
                BroadcasterToSubscriberStatelessMessageType.RESPONSE_NOTIFY
            ):
                raise ValueError

            resp_authorization_length_bytes = await async_read_exact(result.content, 2)
            resp_authorization_length = int.from_bytes(
                resp_authorization_length_bytes, "big", signed=False
            )
            resp_authorization_bytes = await async_read_exact(
                result.content, resp_authorization_length
            )
            resp_authorization = (
                None
                if resp_authorization_bytes == b""
                else resp_authorization_bytes.decode("utf-8")
            )

            resp_tracing_length_bytes = await async_read_exact(result.content, 2)
            resp_tracing_length = int.from_bytes(
                resp_tracing_length_bytes, "big", signed=False
            )
            resp_tracing = await async_read_exact(result.content, resp_tracing_length)

            num_subscribers_bytes = await async_read_exact(result.content, 4)
            num_subscribers = int.from_bytes(num_subscribers_bytes, "big", signed=False)

            resp_identifier_length_bytes = await async_read_exact(result.content, 1)
            resp_identifier_length = int.from_bytes(
                resp_identifier_length_bytes, "big", signed=False
            )
            if resp_identifier_length != len(identifier):
                raise ValueError

            resp_identifier = await async_read_exact(
                result.content, resp_identifier_length
            )
            if resp_identifier != identifier:
                raise ValueError
        except ValueError:
            # also raised by async_read_exact if not enough bytes are read;
            # i don't think this will happen if the connection is interrupted
            # (it should result in a aiohttp.ClientError)
            return BroadcasterRetryableActionResultRetry(
                type=BroadcasterRetryableActionResult.RETRY,
                followup=HttpNotifyActionRetryFollowup(
                    config=self.config,
                    info=self.info,
                    session=self.session,
                    tracer=on_response_auth_result.on_bad_response(),
                    seen_ambiguous=self.seen_ambiguous,
                ),
            )
        except aiohttp.ClientError:
            # connection was interrupted; might have contained a valid
            # response
            return await self._on_ambiguous_retryable(
                retry_tracer=on_response_auth_result.on_bad_response()
            )
        finally:
            await result.__aexit__(None, None, None)

        auth_result = await self.config.is_confirm_notify_allowed(
            tracing=resp_tracing,
            identifier=resp_identifier,
            subscribers=num_subscribers,
            topic=self.info.topic,
            message_sha512=self.info.sha512,
            authorization=resp_authorization,
            now=time.time(),
        )
        if auth_result != AuthResult.OK:
            return BroadcasterRetryableActionResultRetry(
                type=BroadcasterRetryableActionResult.RETRY,
                followup=HttpNotifyActionRetryFollowup(
                    config=self.config,
                    info=self.info,
                    session=self.session,
                    tracer=on_response_auth_result.on_bad_auth_result(
                        result=auth_result
                    ),
                    seen_ambiguous=self.seen_ambiguous,
                ),
            )

        on_response_auth_result.on_response_notify_accepted(
            tracing=resp_tracing, num_subscribers=num_subscribers
        )
        return BroadcasterRetryableActionResultSuccess(
            type=BroadcasterRetryableActionResult.SUCCESS,
            result=HttpPubSubNotifyResult(notified=num_subscribers),
        )

    async def _on_network_error(
        self,
        /,
        *,
        on_response_received_tracer: StatelessTracingNotifyOnResponseReceived,
    ) -> Union[
        BroadcasterRetryableActionResultFailure[Literal["ambiguous", "refused"]],
        BroadcasterRetryableActionResultRetry[
            HttpPubSubNotifyResult, Literal["ambiguous", "refused"]
        ],
    ]:
        return await self._on_ambiguous_retryable(
            retry_tracer=on_response_received_tracer.on_network_error()
        )

    async def _on_ambiguous_retryable(
        self, /, *, retry_tracer: StatelessTracingNotifyOnRetryDetermined
    ) -> Union[
        BroadcasterRetryableActionResultFailure[Literal["ambiguous", "refused"]],
        BroadcasterRetryableActionResultRetry[
            HttpPubSubNotifyResult, Literal["ambiguous", "refused"]
        ],
    ]:
        if self.config.outgoing_retry_ambiguous:
            return BroadcasterRetryableActionResultRetry(
                type=BroadcasterRetryableActionResult.RETRY,
                followup=HttpNotifyActionRetryFollowup(
                    config=self.config,
                    info=self.info,
                    session=self.session,
                    tracer=retry_tracer,
                    seen_ambiguous=True,
                ),
            )

        retry_tracer.on_retry_prevented()
        return BroadcasterRetryableActionResultFailure(
            type=BroadcasterRetryableActionResult.FAILURE,
            result="ambiguous",
        )


class HttpNotifyActionRetryFollowup(Generic[InitializerT]):
    """Retry followup after a request fails in a retryable manner for notify"""

    def __init__(
        self,
        config: HttpPubSubConfig[InitializerT],
        info: HttpNotifyInfo,
        session: aiohttp.ClientSession,
        tracer: StatelessTracingNotifyOnRetryDetermined,
        seen_ambiguous: bool,
    ):
        self.config = config
        self.info = info
        self.session = session
        self.tracer = tracer
        self.seen_ambiguous = seen_ambiguous

    async def on_retries_exhausted(
        self, /
    ) -> BroadcasterRetryableActionResultFailure[Literal["ambiguous", "refused"]]:
        self.tracer.on_retries_exhausted()
        return BroadcasterRetryableActionResultFailure(
            type=BroadcasterRetryableActionResult.FAILURE,
            result="ambiguous" if self.seen_ambiguous else "refused",
        )

    async def on_waiting_to_retry(
        self, /
    ) -> BroadcasterRetryableActionRetryFollowup[
        HttpPubSubNotifyResult, Literal["ambiguous", "refused"]
    ]:
        self.tracer = self.tracer.on_waiting_to_retry()
        return self

    async def on_retrying(
        self, /
    ) -> BroadcasterRetryableAction[
        HttpPubSubNotifyResult, Literal["ambiguous", "refused"]
    ]:
        tracer = self.tracer.on_retrying()
        return HttpNotifyAction(
            config=self.config,
            info=self.info,
            session=self.session,
            tracer=tracer,
            seen_ambiguous=self.seen_ambiguous,
        )


if TYPE_CHECKING:
    _a: Type[
        BroadcasterRetryableAction[
            HttpPubSubNotifyResult, Literal["ambiguous", "refused"]
        ]
    ] = HttpNotifyAction
    _b: Type[
        BroadcasterRetryableActionRetryFollowup[
            HttpPubSubNotifyResult, Literal["ambiguous", "refused"]
        ]
    ] = HttpNotifyActionRetryFollowup



import asyncio
import time

from lonelypsp.stateful.constants import (
    BroadcasterToSubscriberStatefulMessageType,
    SubscriberToBroadcasterStatefulMessageType,
)
from lonelypsp.stateful.messages.confirm_subscribe import (
    B2S_ConfirmSubscribeExact,
    B2S_ConfirmSubscribeGlob,
)
from lonelypsp.stateful.messages.confirm_unsubscribe import (
    B2S_ConfirmUnsubscribeExact,
    B2S_ConfirmUnsubscribeGlob,
)
from lonelypsp.stateful.messages.subscribe import (
    S2B_SubscribeExact,
    S2B_SubscribeGlob,
    serialize_s2b_subscribe_exact,
    serialize_s2b_subscribe_glob,
)
from lonelypsp.stateful.messages.unsubscribe import (
    S2B_UnsubscribeExact,
    S2B_UnsubscribeGlob,
    serialize_s2b_unsubscribe_exact,
    serialize_s2b_unsubscribe_glob,
)

from lonelypsc.ws.check_result import CheckResult
from lonelypsc.ws.handlers.open.websocket_url import (
    make_for_send_websocket_url_and_change_counter,
)
from lonelypsc.ws.state import (
    ManagementTask,
    ManagementTaskType,
    SendingManagementTask,
    SendingState,
    StateOpen,
)


def check_management_tasks(state: StateOpen) -> CheckResult:
    """Checks if the left item from management_tasks can be moved to send_task,
    doing so if possible.

    Critically this does not return successfully if it is possible the management
    task cannot be recovered. Specifically, the management task is immediately
    moved into expected_acks where it can be recovered in cleanup_open, only
    possibly failing with irrecoverable errors

    (Note: cleanup_open using sending.management_task instead would be difficult
    given it would be ambiguous if it had been added to expected_acks or not if it
    was added there in the sending.task)
    """
    if state.sending is not None:
        return CheckResult.CONTINUE

    try:
        task = state.management_tasks.get_nowait()
    except asyncio.QueueEmpty:
        return CheckResult.CONTINUE

    if task.type == ManagementTaskType.SUBSCRIBE_EXACT:
        state.expected_acks.append(
            B2S_ConfirmSubscribeExact(
                type=BroadcasterToSubscriberStatefulMessageType.CONFIRM_SUBSCRIBE_EXACT,
                topic=task.topic,
                tracing=b"",
                authorization=None,
            )
        )
    elif task.type == ManagementTaskType.SUBSCRIBE_GLOB:
        state.expected_acks.append(
            B2S_ConfirmSubscribeGlob(
                type=BroadcasterToSubscriberStatefulMessageType.CONFIRM_SUBSCRIBE_GLOB,
                glob=task.glob,
                tracing=b"",
                authorization=None,
            )
        )
    elif task.type == ManagementTaskType.UNSUBSCRIBE_EXACT:
        state.expected_acks.append(
            B2S_ConfirmUnsubscribeExact(
                type=BroadcasterToSubscriberStatefulMessageType.CONFIRM_UNSUBSCRIBE_EXACT,
                topic=task.topic,
                tracing=b"",
                authorization=None,
            )
        )
    elif task.type == ManagementTaskType.UNSUBSCRIBE_GLOB:
        state.expected_acks.append(
            B2S_ConfirmUnsubscribeGlob(
                type=BroadcasterToSubscriberStatefulMessageType.CONFIRM_UNSUBSCRIBE_GLOB,
                glob=task.glob,
                authorization=None,
                tracing=b"",
            )
        )
    else:
        raise NotImplementedError(f"unknown management task type {task}")

    state.sending = SendingManagementTask(
        type=SendingState.MANAGEMENT_TASK,
        management_task=task,
        task=asyncio.create_task(send_management_task(state, task)),
    )
    return CheckResult.RESTART


async def send_management_task(state: StateOpen, task: ManagementTask) -> None:
    """Target for state.sending.task when using SendingManagementTask; this
    potentially needs multiple event loops to build the authorization value,
    during which time nothing else can send on the websocket since the url
    included in the authorization is order dependent
    """
    url = make_for_send_websocket_url_and_change_counter(state)

    if task.type == ManagementTaskType.SUBSCRIBE_EXACT:
        authorization = await state.config.authorize_subscribe_exact(
            tracing=b"",  # TODO: tracing
            url=url,
            recovery=None,
            exact=task.topic,
            now=time.time(),
        )
        await state.websocket.send_bytes(
            serialize_s2b_subscribe_exact(
                S2B_SubscribeExact(
                    type=SubscriberToBroadcasterStatefulMessageType.SUBSCRIBE_EXACT,
                    authorization=authorization,
                    topic=task.topic,
                    tracing=b"",  # TODO: tracing
                ),
                minimal_headers=state.config.websocket_minimal_headers,
            )
        )
        return

    if task.type == ManagementTaskType.SUBSCRIBE_GLOB:
        authorization = await state.config.authorize_subscribe_glob(
            tracing=b"", url=url, recovery=None, glob=task.glob, now=time.time()
        )
        await state.websocket.send_bytes(
            serialize_s2b_subscribe_glob(
                S2B_SubscribeGlob(
                    type=SubscriberToBroadcasterStatefulMessageType.SUBSCRIBE_GLOB,
                    authorization=authorization,
                    tracing=b"",  # TODO: traicng
                    glob=task.glob,
                ),
                minimal_headers=state.config.websocket_minimal_headers,
            )
        )
        return

    if task.type == ManagementTaskType.UNSUBSCRIBE_EXACT:
        authorization = await state.config.authorize_subscribe_exact(
            tracing=b"", url=url, recovery=None, exact=task.topic, now=time.time()
        )
        await state.websocket.send_bytes(
            serialize_s2b_unsubscribe_exact(
                S2B_UnsubscribeExact(
                    type=SubscriberToBroadcasterStatefulMessageType.UNSUBSCRIBE_EXACT,
                    authorization=authorization,
                    tracing=b"",  # TODO: tracing
                    topic=task.topic,
                ),
                minimal_headers=state.config.websocket_minimal_headers,
            )
        )
        return

    if task.type == ManagementTaskType.UNSUBSCRIBE_GLOB:
        authorization = await state.config.authorize_subscribe_glob(
            tracing=b"", url=url, recovery=None, glob=task.glob, now=time.time()
        )
        await state.websocket.send_bytes(
            serialize_s2b_unsubscribe_glob(
                S2B_UnsubscribeGlob(
                    type=SubscriberToBroadcasterStatefulMessageType.UNSUBSCRIBE_GLOB,
                    authorization=authorization,
                    tracing=b"",  # TODO: tracing
                    glob=task.glob,
                ),
                minimal_headers=state.config.websocket_minimal_headers,
            )
        )
        return

    raise NotImplementedError(f"unknown management task type {task}")



from typing import TYPE_CHECKING

from lonelypsp.stateful.messages.disable_zstd_custom import B2S_DisableZstdCustom

from lonelypsc.client import PubSubError
from lonelypsc.ws.handlers.open.messages.protocol import MessageChecker
from lonelypsc.ws.state import (
    StateOpen,
)


def check_disable_zstd_custom(state: StateOpen, message: B2S_DisableZstdCustom) -> None:
    """Handles the subscriber receiving the a message from the broadcaster that
    it has discarded a custom dictionary it sent earlier, so we can discard it
    as well.

    Raises
    """
    try:
        state.compressors.remove_compressor(message.identifier)
    except KeyError:
        raise PubSubError(f"unknown compressor {message.identifier}")


if TYPE_CHECKING:
    _: MessageChecker[B2S_DisableZstdCustom] = check_disable_zstd_custom



import asyncio
from typing import TYPE_CHECKING

from lonelypsp.stateful.messages.enable_zstd_custom import B2S_EnableZstdCustom

from lonelypsc.client import PubSubError
from lonelypsc.ws.compressor import (
    CompressorPreparing,
    CompressorReady,
    CompressorState,
)
from lonelypsc.ws.handlers.open.messages.protocol import MessageChecker
from lonelypsc.ws.state import (
    StateOpen,
)

try:
    import zstandard
except ImportError:
    ...


def _make_compressor(message: B2S_EnableZstdCustom) -> CompressorReady:
    zdict = zstandard.ZstdCompressionDict(message.dictionary)
    zdict.precompute_compress(level=message.compression_level)
    return CompressorReady(
        type=CompressorState.READY,
        identifier=message.identifier,
        level=message.compression_level,
        min_size=message.min_size,
        max_size=message.max_size,
        data=zdict,
        compressors=list(),
        decompressors=list(),
    )


def check_enable_zstd_custom(state: StateOpen, message: B2S_EnableZstdCustom) -> None:
    """Handles the subscriber receiving the a message from the broadcaster that
    the broadcaster has created a dictionary for compressing the types of
    messages that have been sent over this websocket
    """
    if not state.config.allow_compression:
        raise PubSubError("compression is disabled but a dictionary was received")
    if not state.config.allow_training_compression:
        raise PubSubError(
            "compression training is disabled but a dictionary was received"
        )

    state.compressors.add_compressor(
        CompressorPreparing(
            type=CompressorState.PREPARING,
            identifier=message.identifier,
            task=asyncio.create_task(asyncio.to_thread(_make_compressor, message)),
        )
    )


if TYPE_CHECKING:
    _: MessageChecker[B2S_EnableZstdCustom] = check_enable_zstd_custom



from typing import (
    TYPE_CHECKING,
    Generic,
    List,
    Literal,
    Optional,
    Protocol,
    Type,
    TypedDict,
    TypeVar,
    Union,
)

from fastapi import APIRouter
from lonelypsp.auth.config import AuthConfig, AuthResult
from lonelypsp.auth.set_subscriptions_info import SetSubscriptionsInfo
from lonelypsp.stateful.messages.configure import S2B_Configure
from lonelypsp.stateful.messages.confirm_configure import B2S_ConfirmConfigure
from lonelypsp.stateful.messages.continue_notify import B2S_ContinueNotify
from lonelypsp.stateful.messages.continue_receive import S2B_ContinueReceive
from lonelypsp.stateful.messages.disable_zstd_custom import B2S_DisableZstdCustom
from lonelypsp.stateful.messages.enable_zstd_custom import B2S_EnableZstdCustom
from lonelypsp.stateful.messages.enable_zstd_preset import B2S_EnableZstdPreset
from lonelypsp.stateless.make_strong_etag import StrongEtag
from lonelypsp.tracing.stateless.root import (
    InitializerTcontra,
    StatelessTracingSubscriberRoot,
)

from lonelypsc.config.config import PubSubBroadcasterConfig


class HttpPubSubBindUvicornConfig(TypedDict):
    """When used for the `bind` parameter on an `HttpPubSubConfig` object,
    indicates you want to use a relatively standard http server with uvicorn.
    This is provided as the most common option, but if you need to configure
    anything more specific (e.g., middleware for logging, etc), you should
    use `HttpPubSubBindManualConfig` instead.

    Note - this is just converted into a manual config via the module
    `lonelypsc.config.helpers.uvicorn_bind_config`
    """

    type: Literal["uvicorn"]
    """discriminates the type"""

    host: str
    """What address to bind to, e.g., 0.0.0.0 for all interfaces"""

    port: int
    """What port to bind to. As a very loose guidance, use 3002 for subscribers 
    and 3003 for broadcasters
    """


class HttpPubSubBindManualCallback(Protocol):
    """Describes the callback that binds the http server in manual mode"""

    async def __call__(self, router: APIRouter) -> None:
        """Serves requests continuously with the given router, not returning
        until the server is shutdown. We will cancel the task when the client
        is being exitted. Should never return normally, as then there would be
        no way to know when to shut down the server. The only exception is if
        you are guarranteeing only one client in the lifetime of the application.
        """


class HttpPubSubBindManualConfig(TypedDict):
    """When used for the `bind` parameter on an `HttpPubSubConfig` object,
    indicates you can convert the APIRouter into an HTTP server yourself.
    Often its convenient to use this just because you want to provide more
    parameters to uvicorn (e.g., TLS), or to add middleware, etc.
    """

    type: Literal["manual"]
    """discriminates the type"""

    callback: HttpPubSubBindManualCallback
    """The callback that binds the http server."""


class HttpPubSubBindConfig(Protocol):
    """Determines how the broadcaster can reach us"""

    @property
    def bind(self) -> Union[HttpPubSubBindUvicornConfig, HttpPubSubBindManualConfig]:
        """Determines how the FastAPI APIRouter is converted into an HTTP server"""

    @property
    def host(self) -> str:
        """The schema and address that the broadcaster should use to reach us.

        This CAN include a path component, which is assumed to be a prefix for
        the standard routes, and a fragment, which is kept (but potentially
        appended to) for our subscriptions. When using multiple
        lonelypscs at the same schema and adddress they need to be
        distinguished either by path or fragment

        WARN:
            This information is _not_ incorporated into the APIRouter that is passed
            to the bind config. In other words, if the APIRouter is served as is
            (like the "uvicorn" strategy does), then you will be expecting to receive
            requests at the root path, which means that a proxy between the broadcaster
            and us must be rewriting requests to the correct path.

            If you want to use a prefix path but do not want to use a rewriting proxy,
            you can use the "manual" strategy and provide the appropriate arguments to
            include_router, e.g., `app.include_router(router, prefix="/myprefix")`.
        """


class HttpPubSubBindConfigFromParts:
    """Implementation of HttpPubSubBindConfig from the various parts"""

    def __init__(
        self,
        /,
        *,
        bind: Union[HttpPubSubBindUvicornConfig, HttpPubSubBindManualConfig],
        host: str,
    ):
        self.bind = bind
        self.host = host


if TYPE_CHECKING:
    _: Type[HttpPubSubBindConfig] = HttpPubSubBindConfigFromParts


class HttpPubSubConnectConfig(Protocol):
    """Determines how we connect to broadcasters"""

    @property
    def broadcasters(self) -> List[PubSubBroadcasterConfig]:
        """The broadcasters that we can connect to for making subscription requests
        and requesting a message be broadcast
        """

    @property
    def outgoing_retries_per_broadcaster(self) -> int:
        """How many times to retry a broadcaster before giving up"""


class HttpPubSubConnectConfigFromParts:
    """Implementation of HttpPubSubConnectConfig from the various parts"""

    def __init__(
        self,
        /,
        *,
        broadcasters: List[PubSubBroadcasterConfig],
        outgoing_retries_per_broadcaster: int,
    ):
        self.broadcasters = broadcasters
        self.outgoing_retries_per_broadcaster = outgoing_retries_per_broadcaster


if TYPE_CHECKING:
    __: Type[HttpPubSubConnectConfig] = HttpPubSubConnectConfigFromParts


class HttpPubSubGenericConfig(Protocol):
    """Generic network configuration."""

    @property
    def message_body_spool_size(self) -> int:
        """If the message body exceeds this size we always switch to a temporary file."""

    @property
    def outgoing_http_timeout_total(self) -> Optional[float]:
        """The total timeout for outgoing (to broadcaster) http requests in seconds"""

    @property
    def outgoing_http_timeout_connect(self) -> Optional[float]:
        """The timeout for connecting to the broadcaster in seconds, which may include multiple
        socket attempts
        """

    @property
    def outgoing_http_timeout_sock_read(self) -> Optional[float]:
        """The timeout for reading from a broadcaster socket in seconds before the socket is
        considered dead
        """

    @property
    def outgoing_http_timeout_sock_connect(self) -> Optional[float]:
        """The timeout for a single socket connecting to a broadcaster before we give up, in seconds"""

    @property
    def outgoing_retry_ambiguous(self) -> bool:
        """Determines how to handle when we are unsure if a broadcaster received a message
        and the message is not necessarily idempotent.

        - **True** _(recommended)_: if we are unsure, we will assume the broadcaster
          did NOT receive the message.

        - **False** _(not recommended)_: if we are unsure, we will assume the
          broadcaster DID receive the message.

        Note that this scenario is theoretically guarranteed to be possible;
        this is known as the Two Generals' problem. However, the client may
        treat some scenarios as ambiguous where it could be known it wasn't
        received with a lower-level inspection of the packets.

        Thus, this is a tradeoff between duplicating messages when the guess is
        wrong (True) and dropping messages when the guess is wrong (False).
        """


class HttpPubSubGenericConfigFromParts:
    """Convenience class that allows you to create a GenericConfig protocol
    satisfying object from parts"""

    def __init__(
        self,
        message_body_spool_size: int,
        outgoing_http_timeout_total: Optional[float],
        outgoing_http_timeout_connect: Optional[float],
        outgoing_http_timeout_sock_read: Optional[float],
        outgoing_http_timeout_sock_connect: Optional[float],
        outgoing_retry_ambiguous: bool,
    ):
        self.message_body_spool_size = message_body_spool_size
        self.outgoing_http_timeout_total = outgoing_http_timeout_total
        self.outgoing_http_timeout_connect = outgoing_http_timeout_connect
        self.outgoing_http_timeout_sock_read = outgoing_http_timeout_sock_read
        self.outgoing_http_timeout_sock_connect = outgoing_http_timeout_sock_connect
        self.outgoing_retry_ambiguous = outgoing_retry_ambiguous


if TYPE_CHECKING:
    ___: Type[HttpPubSubGenericConfig] = HttpPubSubGenericConfigFromParts


InitializerT = TypeVar("InitializerT")
InitializerTco = TypeVar("InitializerTco", covariant=True)


class TracingConfig(Generic[InitializerTcontra], Protocol):
    @property
    def tracing(self) -> StatelessTracingSubscriberRoot[InitializerTcontra]:
        """Generates tracing objects for the various requests when given some
        initializer data
        """


class HttpPubSubConfig(
    HttpPubSubBindConfig,
    HttpPubSubConnectConfig,
    HttpPubSubGenericConfig,
    AuthConfig,
    TracingConfig[InitializerTcontra],
    Generic[InitializerTcontra],
    Protocol,
):
    """Configures the library, including how we are reached, how we connect
    to broadcasters, and how we provide/verify authorization
    """


class HttpPubSubConfigFromParts(Generic[InitializerT]):
    """An implementation of HttpPubSubConfig from the various parts"""

    def __init__(
        self,
        /,
        *,
        bind_config: HttpPubSubBindConfig,
        connect_config: HttpPubSubConnectConfig,
        generic_config: HttpPubSubGenericConfig,
        auth_config: AuthConfig,
        tracing: StatelessTracingSubscriberRoot[InitializerT],
    ):
        self.bind_config = bind_config
        self.connect_config = connect_config
        self.generic_config = generic_config
        self.auth = auth_config
        self.tracing = tracing

    @property
    def bind(self) -> Union[HttpPubSubBindUvicornConfig, HttpPubSubBindManualConfig]:
        return self.bind_config.bind

    @property
    def host(self) -> str:
        return self.bind_config.host

    @property
    def broadcasters(self) -> List[PubSubBroadcasterConfig]:
        return self.connect_config.broadcasters

    @property
    def outgoing_retries_per_broadcaster(self) -> int:
        return self.connect_config.outgoing_retries_per_broadcaster

    @property
    def message_body_spool_size(self) -> int:
        return self.generic_config.message_body_spool_size

    @property
    def outgoing_http_timeout_total(self) -> Optional[float]:
        return self.generic_config.outgoing_http_timeout_total

    @property
    def outgoing_http_timeout_connect(self) -> Optional[float]:
        return self.generic_config.outgoing_http_timeout_connect

    @property
    def outgoing_http_timeout_sock_read(self) -> Optional[float]:
        return self.generic_config.outgoing_http_timeout_sock_read

    @property
    def outgoing_http_timeout_sock_connect(self) -> Optional[float]:
        return self.generic_config.outgoing_http_timeout_sock_connect

    @property
    def outgoing_retry_ambiguous(self) -> bool:
        return self.generic_config.outgoing_retry_ambiguous

    async def setup_to_broadcaster_auth(self) -> None:
        await self.auth.setup_to_broadcaster_auth()

    async def teardown_to_broadcaster_auth(self) -> None:
        await self.auth.teardown_to_broadcaster_auth()

    async def setup_to_subscriber_auth(self) -> None:
        await self.auth.setup_to_subscriber_auth()

    async def teardown_to_subscriber_auth(self) -> None:
        await self.auth.teardown_to_subscriber_auth()

    async def authorize_subscribe_exact(
        self,
        /,
        *,
        tracing: bytes,
        url: str,
        recovery: Optional[str],
        exact: bytes,
        now: float,
    ) -> Optional[str]:
        return await self.auth.authorize_subscribe_exact(
            tracing=tracing, url=url, recovery=recovery, exact=exact, now=now
        )

    async def is_subscribe_exact_allowed(
        self,
        /,
        *,
        tracing: bytes,
        url: str,
        recovery: Optional[str],
        exact: bytes,
        now: float,
        authorization: Optional[str],
    ) -> AuthResult:
        return await self.auth.is_subscribe_exact_allowed(
            tracing=tracing,
            url=url,
            recovery=recovery,
            exact=exact,
            now=now,
            authorization=authorization,
        )

    async def authorize_subscribe_glob(
        self,
        /,
        *,
        tracing: bytes,
        url: str,
        recovery: Optional[str],
        glob: str,
        now: float,
    ) -> Optional[str]:
        return await self.auth.authorize_subscribe_glob(
            tracing=tracing, url=url, recovery=recovery, glob=glob, now=now
        )

    async def is_subscribe_glob_allowed(
        self,
        /,
        *,
        tracing: bytes,
        url: str,
        recovery: Optional[str],
        glob: str,
        now: float,
        authorization: Optional[str],
    ) -> AuthResult:
        return await self.auth.is_subscribe_glob_allowed(
            tracing=tracing,
            url=url,
            recovery=recovery,
            glob=glob,
            now=now,
            authorization=authorization,
        )

    async def authorize_notify(
        self,
        /,
        *,
        tracing: bytes,
        topic: bytes,
        identifier: bytes,
        message_sha512: bytes,
        now: float,
    ) -> Optional[str]:
        return await self.auth.authorize_notify(
            tracing=tracing,
            topic=topic,
            identifier=identifier,
            message_sha512=message_sha512,
            now=now,
        )

    async def is_notify_allowed(
        self,
        /,
        *,
        tracing: bytes,
        topic: bytes,
        identifier: bytes,
        message_sha512: bytes,
        now: float,
        authorization: Optional[str],
    ) -> AuthResult:
        return await self.auth.is_notify_allowed(
            tracing=tracing,
            topic=topic,
            identifier=identifier,
            message_sha512=message_sha512,
            now=now,
            authorization=authorization,
        )

    async def authorize_stateful_configure(
        self,
        /,
        *,
        tracing: bytes,
        subscriber_nonce: bytes,
        enable_zstd: bool,
        enable_training: bool,
        initial_dict: int,
    ) -> Optional[str]:
        return await self.auth.authorize_stateful_configure(
            tracing=tracing,
            subscriber_nonce=subscriber_nonce,
            enable_zstd=enable_zstd,
            enable_training=enable_training,
            initial_dict=initial_dict,
        )

    async def is_stateful_configure_allowed(
        self, /, *, message: S2B_Configure, now: float
    ) -> AuthResult:
        return await self.auth.is_stateful_configure_allowed(message=message, now=now)

    async def authorize_check_subscriptions(
        self, /, *, tracing: bytes, url: str, now: float
    ) -> Optional[str]:
        return await self.auth.authorize_check_subscriptions(
            tracing=tracing, url=url, now=now
        )

    async def is_check_subscriptions_allowed(
        self, /, *, tracing: bytes, url: str, now: float, authorization: Optional[str]
    ) -> AuthResult:
        return await self.auth.is_check_subscriptions_allowed(
            tracing=tracing, url=url, now=now, authorization=authorization
        )

    async def authorize_set_subscriptions(
        self, /, *, tracing: bytes, url: str, strong_etag: StrongEtag, now: float
    ) -> Optional[str]:
        return await self.auth.authorize_set_subscriptions(
            tracing=tracing, url=url, strong_etag=strong_etag, now=now
        )

    async def is_set_subscriptions_allowed(
        self,
        /,
        *,
        tracing: bytes,
        url: str,
        strong_etag: StrongEtag,
        subscriptions: SetSubscriptionsInfo,
        now: float,
        authorization: Optional[str],
    ) -> AuthResult:
        return await self.auth.is_set_subscriptions_allowed(
            tracing=tracing,
            url=url,
            strong_etag=strong_etag,
            subscriptions=subscriptions,
            now=now,
            authorization=authorization,
        )

    async def authorize_stateful_continue_receive(
        self,
        /,
        *,
        tracing: bytes,
        identifier: bytes,
        part_id: int,
        url: str,
        now: float,
    ) -> Optional[str]:
        return await self.auth.authorize_stateful_continue_receive(
            tracing=tracing, identifier=identifier, part_id=part_id, url=url, now=now
        )

    async def is_stateful_continue_receive_allowed(
        self, /, *, url: str, message: S2B_ContinueReceive, now: float
    ) -> AuthResult:
        return await self.auth.is_stateful_continue_receive_allowed(
            url=url, message=message, now=now
        )

    async def authorize_confirm_receive(
        self,
        /,
        *,
        tracing: bytes,
        identifier: bytes,
        num_subscribers: int,
        url: str,
        now: float,
    ) -> Optional[str]:
        return await self.auth.authorize_confirm_receive(
            tracing=tracing,
            identifier=identifier,
            num_subscribers=num_subscribers,
            url=url,
            now=now,
        )

    async def is_confirm_receive_allowed(
        self,
        /,
        *,
        tracing: bytes,
        identifier: bytes,
        num_subscribers: int,
        url: str,
        now: float,
        authorization: Optional[str],
    ) -> AuthResult:
        return await self.auth.is_confirm_receive_allowed(
            tracing=tracing,
            identifier=identifier,
            num_subscribers=num_subscribers,
            url=url,
            now=now,
            authorization=authorization,
        )

    async def authorize_confirm_missed(
        self, /, *, tracing: bytes, topic: bytes, url: str, now: float
    ) -> Optional[str]:
        return await self.auth.authorize_confirm_missed(
            tracing=tracing, topic=topic, url=url, now=now
        )

    async def is_confirm_missed_allowed(
        self,
        /,
        *,
        tracing: bytes,
        topic: bytes,
        url: str,
        now: float,
        authorization: Optional[str],
    ) -> AuthResult:
        return await self.auth.is_confirm_missed_allowed(
            tracing=tracing, topic=topic, url=url, now=now, authorization=authorization
        )

    async def authorize_receive(
        self,
        /,
        *,
        tracing: bytes,
        url: str,
        topic: bytes,
        message_sha512: bytes,
        identifier: bytes,
        now: float,
    ) -> Optional[str]:
        return await self.auth.authorize_receive(
            tracing=tracing,
            url=url,
            topic=topic,
            message_sha512=message_sha512,
            identifier=identifier,
            now=now,
        )

    async def is_receive_allowed(
        self,
        /,
        *,
        tracing: bytes,
        url: str,
        topic: bytes,
        message_sha512: bytes,
        identifier: bytes,
        now: float,
        authorization: Optional[str],
    ) -> AuthResult:
        return await self.auth.is_receive_allowed(
            tracing=tracing,
            url=url,
            topic=topic,
            message_sha512=message_sha512,
            identifier=identifier,
            now=now,
            authorization=authorization,
        )

    async def authorize_missed(
        self, /, *, tracing: bytes, recovery: str, topic: bytes, now: float
    ) -> Optional[str]:
        return await self.auth.authorize_missed(
            tracing=tracing, recovery=recovery, topic=topic, now=now
        )

    async def is_missed_allowed(
        self,
        /,
        *,
        tracing: bytes,
        recovery: str,
        topic: bytes,
        now: float,
        authorization: Optional[str],
    ) -> AuthResult:
        return await self.auth.is_missed_allowed(
            tracing=tracing,
            recovery=recovery,
            topic=topic,
            now=now,
            authorization=authorization,
        )

    async def authorize_confirm_subscribe_exact(
        self,
        /,
        *,
        tracing: bytes,
        url: str,
        recovery: Optional[str],
        exact: bytes,
        now: float,
    ) -> Optional[str]:
        return await self.auth.authorize_confirm_subscribe_exact(
            tracing=tracing, url=url, recovery=recovery, exact=exact, now=now
        )

    async def is_confirm_subscribe_exact_allowed(
        self,
        /,
        *,
        tracing: bytes,
        url: str,
        recovery: Optional[str],
        exact: bytes,
        now: float,
        authorization: Optional[str],
    ) -> AuthResult:
        return await self.auth.is_confirm_subscribe_exact_allowed(
            tracing=tracing,
            url=url,
            recovery=recovery,
            exact=exact,
            now=now,
            authorization=authorization,
        )

    async def authorize_confirm_subscribe_glob(
        self,
        /,
        *,
        tracing: bytes,
        url: str,
        recovery: Optional[str],
        glob: str,
        now: float,
    ) -> Optional[str]:
        return await self.auth.authorize_confirm_subscribe_glob(
            tracing=tracing, url=url, recovery=recovery, glob=glob, now=now
        )

    async def is_confirm_subscribe_glob_allowed(
        self,
        /,
        *,
        tracing: bytes,
        url: str,
        recovery: Optional[str],
        glob: str,
        now: float,
        authorization: Optional[str],
    ) -> AuthResult:
        return await self.auth.is_confirm_subscribe_glob_allowed(
            tracing=tracing,
            url=url,
            recovery=recovery,
            glob=glob,
            now=now,
            authorization=authorization,
        )

    async def authorize_confirm_notify(
        self,
        /,
        *,
        tracing: bytes,
        identifier: bytes,
        subscribers: int,
        topic: bytes,
        message_sha512: bytes,
        now: float,
    ) -> Optional[str]:
        return await self.auth.authorize_confirm_notify(
            tracing=tracing,
            identifier=identifier,
            subscribers=subscribers,
            topic=topic,
            message_sha512=message_sha512,
            now=now,
        )

    async def is_confirm_notify_allowed(
        self,
        /,
        *,
        tracing: bytes,
        identifier: bytes,
        subscribers: int,
        topic: bytes,
        message_sha512: bytes,
        authorization: Optional[str],
        now: float,
    ) -> AuthResult:
        return await self.auth.is_confirm_notify_allowed(
            tracing=tracing,
            identifier=identifier,
            subscribers=subscribers,
            topic=topic,
            message_sha512=message_sha512,
            authorization=authorization,
            now=now,
        )

    async def authorize_check_subscriptions_response(
        self,
        /,
        *,
        tracing: bytes,
        strong_etag: StrongEtag,
        now: float,
    ) -> Optional[str]:
        return await self.auth.authorize_check_subscriptions_response(
            tracing=tracing, strong_etag=strong_etag, now=now
        )

    async def is_check_subscription_response_allowed(
        self,
        /,
        *,
        tracing: bytes,
        strong_etag: StrongEtag,
        authorization: Optional[str],
        now: float,
    ) -> AuthResult:
        return await self.auth.is_check_subscription_response_allowed(
            tracing=tracing,
            strong_etag=strong_etag,
            authorization=authorization,
            now=now,
        )

    async def authorize_set_subscriptions_response(
        self, /, *, tracing: bytes, strong_etag: StrongEtag, now: float
    ) -> Optional[str]:
        return await self.auth.authorize_set_subscriptions_response(
            tracing=tracing, strong_etag=strong_etag, now=now
        )

    async def is_set_subscription_response_allowed(
        self,
        /,
        *,
        tracing: bytes,
        strong_etag: StrongEtag,
        authorization: Optional[str],
        now: float,
    ) -> AuthResult:
        return await self.auth.is_set_subscription_response_allowed(
            tracing=tracing,
            strong_etag=strong_etag,
            authorization=authorization,
            now=now,
        )

    async def authorize_stateful_confirm_configure(
        self, /, *, broadcaster_nonce: bytes, tracing: bytes, now: float
    ) -> Optional[str]:
        return await self.auth.authorize_stateful_confirm_configure(
            broadcaster_nonce=broadcaster_nonce, tracing=tracing, now=now
        )

    async def is_stateful_confirm_configure_allowed(
        self, /, *, message: B2S_ConfirmConfigure, now: float
    ) -> AuthResult:
        return await self.auth.is_stateful_confirm_configure_allowed(
            message=message, now=now
        )

    async def authorize_stateful_enable_zstd_preset(
        self,
        /,
        *,
        tracing: bytes,
        url: str,
        compressor_identifier: int,
        compression_level: int,
        min_size: int,
        max_size: int,
        now: float,
    ) -> Optional[str]:
        return await self.auth.authorize_stateful_enable_zstd_preset(
            tracing=tracing,
            url=url,
            compressor_identifier=compressor_identifier,
            compression_level=compression_level,
            min_size=min_size,
            max_size=max_size,
            now=now,
        )

    async def is_stateful_enable_zstd_preset_allowed(
        self, /, *, url: str, message: B2S_EnableZstdPreset, now: float
    ) -> AuthResult:
        return await self.auth.is_stateful_enable_zstd_preset_allowed(
            url=url, message=message, now=now
        )

    async def authorize_stateful_enable_zstd_custom(
        self,
        /,
        *,
        tracing: bytes,
        url: str,
        compressor_identifier: int,
        compression_level: int,
        min_size: int,
        max_size: int,
        sha512: bytes,
        now: float,
    ) -> Optional[str]:
        return await self.auth.authorize_stateful_enable_zstd_custom(
            tracing=tracing,
            url=url,
            compressor_identifier=compressor_identifier,
            compression_level=compression_level,
            min_size=min_size,
            max_size=max_size,
            sha512=sha512,
            now=now,
        )

    async def is_stateful_enable_zstd_custom_allowed(
        self, /, *, url: str, message: B2S_EnableZstdCustom, now: float
    ) -> AuthResult:
        return await self.auth.is_stateful_enable_zstd_custom_allowed(
            url=url, message=message, now=now
        )

    async def authorize_stateful_disable_zstd_custom(
        self, /, *, tracing: bytes, compressor_identifier: int, url: str, now: float
    ) -> Optional[str]:
        return await self.auth.authorize_stateful_disable_zstd_custom(
            tracing=tracing,
            compressor_identifier=compressor_identifier,
            url=url,
            now=now,
        )

    async def is_stateful_disable_zstd_custom_allowed(
        self, /, *, url: str, message: B2S_DisableZstdCustom, now: float
    ) -> AuthResult:
        return await self.auth.is_stateful_disable_zstd_custom_allowed(
            url=url, message=message, now=now
        )

    async def authorize_stateful_continue_notify(
        self, /, *, tracing: bytes, identifier: bytes, part_id: int, now: float
    ) -> Optional[str]:
        return await self.auth.authorize_stateful_continue_notify(
            tracing=tracing, identifier=identifier, part_id=part_id, now=now
        )

    async def is_stateful_continue_notify_allowed(
        self, /, *, message: B2S_ContinueNotify, now: float
    ) -> AuthResult:
        return await self.auth.is_stateful_continue_notify_allowed(
            message=message, now=now
        )


if TYPE_CHECKING:
    ____: Type[HttpPubSubConfig] = HttpPubSubConfigFromParts


def make_http_pub_sub_config(
    *,
    bind: Union[HttpPubSubBindUvicornConfig, HttpPubSubBindManualConfig],
    host: str,
    broadcasters: List[PubSubBroadcasterConfig],
    outgoing_retries_per_broadcaster: int,
    message_body_spool_size: int,
    outgoing_http_timeout_total: Optional[float],
    outgoing_http_timeout_connect: Optional[float],
    outgoing_http_timeout_sock_read: Optional[float],
    outgoing_http_timeout_sock_connect: Optional[float],
    outgoing_retry_ambiguous: bool,
    auth: AuthConfig,
    tracing: StatelessTracingSubscriberRoot[InitializerT],
) -> HttpPubSubConfig[InitializerT]:
    """Convenience function to make a HttpPubSubConfig object without excessive nesting
    if you are specifying everything that doesn't need to be synced with the broadcaster
    within code.
    """
    return HttpPubSubConfigFromParts(
        bind_config=HttpPubSubBindConfigFromParts(bind=bind, host=host),
        connect_config=HttpPubSubConnectConfigFromParts(
            broadcasters=broadcasters,
            outgoing_retries_per_broadcaster=outgoing_retries_per_broadcaster,
        ),
        generic_config=HttpPubSubGenericConfigFromParts(
            message_body_spool_size=message_body_spool_size,
            outgoing_http_timeout_total=outgoing_http_timeout_total,
            outgoing_http_timeout_connect=outgoing_http_timeout_connect,
            outgoing_http_timeout_sock_read=outgoing_http_timeout_sock_read,
            outgoing_http_timeout_sock_connect=outgoing_http_timeout_sock_connect,
            outgoing_retry_ambiguous=outgoing_retry_ambiguous,
        ),
        auth_config=auth,
        tracing=tracing,
    )



from lonelypsc.client import PubSubError
from lonelypsc.ws.check_result import CheckResult
from lonelypsc.ws.internal_callbacks import sweep_internal_message
from lonelypsc.ws.state import SendingState, StateOpen


def check_sending(state: StateOpen) -> CheckResult:
    """Checks if sending is done, and if so, verifies it was successful
    then sets it to None, otherwise raises an exception
    """
    if state.sending is None or not state.sending.task.done():
        return CheckResult.CONTINUE

    if state.sending.task.exception() is not None:
        # let cleanup be handled in cleanup_open, which will discover
        # the error again, so avoid having it repeated in the stack trace
        raise PubSubError("send failed")

    if state.sending.type == SendingState.INTERNAL_MESSAGE:
        sweep_internal_message(state.sending.internal_message)

    state.sending = None

    return CheckResult.RESTART



from typing import Protocol, Union


class AsyncReadableBytesIOA(Protocol):
    """A type that represents a stream that can be read synchronously"""

    async def read(self, n: int) -> bytes:
        """Reads n bytes from the file-like object"""
        raise NotImplementedError()


class AsyncReadableBytesIOB(Protocol):
    """A type that represents a stream that can be read synchronously"""

    async def read(self, n: int, /) -> bytes:
        """Reads n bytes from the file-like object"""
        raise NotImplementedError()


AsyncReadableBytesIO = Union[AsyncReadableBytesIOA, AsyncReadableBytesIOB]


async def async_read_exact(stream: AsyncReadableBytesIO, n: int) -> bytes:
    """Reads exactly n bytes from the stream, otherwise raises ValueError"""
    result = await stream.read(n)
    if len(result) != n:
        raise ValueError(f"expected {n} bytes, got {len(result)}")
    return result



import time

from lonelypsp.stateful.constants import SubscriberToBroadcasterStatefulMessageType
from lonelypsp.stateful.messages.confirm_receive import (
    S2B_ConfirmReceive,
    serialize_s2b_confirm_receive,
)

from lonelypsc.ws.handlers.open.websocket_url import (
    make_for_send_websocket_url_and_change_counter,
)
from lonelypsc.ws.state import StateOpen, UnsentAckConfirmReceive


async def send_confirm_receive(
    state: StateOpen, message: UnsentAckConfirmReceive
) -> None:
    """Produces the required authorization header and sends the confirm receive message"""

    url = make_for_send_websocket_url_and_change_counter(state)
    tracing = b""  # TODO: tracing
    num_subscribers = 1
    authorization = await state.config.authorize_confirm_receive(
        tracing=tracing,
        identifier=message.identifier,
        num_subscribers=num_subscribers,
        url=url,
        now=time.time(),
    )
    serd_message = serialize_s2b_confirm_receive(
        S2B_ConfirmReceive(
            type=SubscriberToBroadcasterStatefulMessageType.CONFIRM_RECEIVE,
            identifier=message.identifier,
            authorization=authorization,
            tracing=tracing,
            num_subscribers=num_subscribers,
        ),
        minimal_headers=state.config.websocket_minimal_headers,
    )
    await state.websocket.send_bytes(serd_message)



from typing import Generic, Protocol, TypeVar

from lonelypsp.stateful.message import B2S_Message

from lonelypsc.ws.state import StateOpen

T = TypeVar("T", bound=B2S_Message, contravariant=True)


class MessageChecker(Generic[T], Protocol):
    """Protocol to ensure that we are being consistent with how functions within
    this folder are defined for handling messages received from the broadcaster
    """

    def __call__(self, state: StateOpen, message: T) -> None: ...



import asyncio
import random
from enum import Enum, auto
from typing import Generic, Literal, Optional, Protocol, TypeVar, Union, cast

from lonelypsp.compat import assert_never, fast_dataclass

from lonelypsc.config.config import BroadcastersShuffler, PubSubBroadcasterConfig
from lonelypsc.config.http_config import HttpPubSubConfig

SuccessT = TypeVar("SuccessT")
FailureT = TypeVar("FailureT")
InitializerT = TypeVar("InitializerT")


class BroadcasterRetryableActionResult(Enum):
    SUCCESS = auto()
    """don't retry: have a SuccessT"""

    FAILURE = auto()
    """don't retry: have a FailureT"""

    RETRY = auto()
    """prefer to retry but otherwise have a FailureT"""


@fast_dataclass
class BroadcasterRetryableActionResultSuccess(Generic[SuccessT]):
    type: Literal[BroadcasterRetryableActionResult.SUCCESS]
    """discriminator value"""

    result: SuccessT
    """the result of the action"""


@fast_dataclass
class BroadcasterRetryableActionResultFailure(Generic[FailureT]):
    type: Literal[BroadcasterRetryableActionResult.FAILURE]
    """discriminator value"""

    result: FailureT
    """the result of the action"""


@fast_dataclass
class BroadcasterRetryableActionResultRetry(Generic[SuccessT, FailureT]):
    type: Literal[BroadcasterRetryableActionResult.RETRY]
    """discriminator value"""

    followup: "BroadcasterRetryableActionRetryFollowup[SuccessT, FailureT]"
    """the callback once the next course of action is determined"""


class BroadcasterRetryableAction(Generic[SuccessT, FailureT], Protocol):
    """Describes some action that performed against a broadcaster
    by the http connector
    """

    async def attempt(self, /, *, broadcaster: PubSubBroadcasterConfig) -> Union[
        BroadcasterRetryableActionResultSuccess[SuccessT],
        BroadcasterRetryableActionResultFailure[FailureT],
        BroadcasterRetryableActionResultRetry[SuccessT, FailureT],
    ]:
        """Attempts the given broadcaster. Only called once per logical
        instance (ie., if it is called again, it means this returned retry
        and the retry function returned this instance again)
        """


class BroadcasterRetryableActionRetryFollowup(Generic[SuccessT, FailureT], Protocol):
    async def on_retries_exhausted(
        self, /
    ) -> BroadcasterRetryableActionResultFailure[FailureT]: ...

    async def on_waiting_to_retry(
        self, /
    ) -> "BroadcasterRetryableActionRetryFollowup[SuccessT, FailureT]": ...

    async def on_retrying(
        self, /
    ) -> "BroadcasterRetryableAction[SuccessT, FailureT]": ...


async def attempt_broadcasters(
    *,
    config: HttpPubSubConfig[InitializerT],
    shuffler: BroadcastersShuffler,
    action: BroadcasterRetryableAction[SuccessT, FailureT],
) -> Union[
    BroadcasterRetryableActionResultSuccess[SuccessT],
    BroadcasterRetryableActionResultFailure[FailureT],
]:
    """Attempts broadcasters one at a time until all retries are exhausted
    or a success is found, with callbacks for all the relevant points so that
    tracing data can be collected.
    """

    next_action = cast(Optional[BroadcasterRetryableAction[SuccessT, FailureT]], action)
    next_retry = cast(
        Optional[BroadcasterRetryableActionRetryFollowup[SuccessT, FailureT]], None
    )
    del action

    for iteration in range(config.outgoing_retries_per_broadcaster):
        if iteration > 0:
            assert next_retry is not None
            next_retry = await next_retry.on_waiting_to_retry()
            await asyncio.sleep(2 ** (iteration - 1) + random.random())

        for broadcaster in shuffler:
            if next_retry is not None:
                next_action = await next_retry.on_retrying()
                next_retry = None

            assert next_action is not None
            result = await next_action.attempt(broadcaster=broadcaster)
            if result.type == BroadcasterRetryableActionResult.SUCCESS:
                return result
            if result.type == BroadcasterRetryableActionResult.FAILURE:
                return result
            if result.type != BroadcasterRetryableActionResult.RETRY:
                assert_never(result)

            next_retry = result.followup
            next_action = None

    if next_action is None:
        raise Exception("no broadcasters to try!")

    assert next_retry is not None
    return await next_retry.on_retries_exhausted()



import json
from typing import Literal, Optional, Tuple, cast

from lonelypsp.auth.config import ToBroadcasterAuthConfig, ToSubscriberAuthConfig
from lonelypsp.auth.helpers.hmac_auth_config import (
    IncomingHmacAuthDBConfig,
    IncomingHmacAuthDBReentrantConfig,
    IncomingHmacAuthSqliteDBConfig,
    ToBroadcasterHmacAuth,
    ToSubscriberHmacAuth,
)
from lonelypsp.auth.helpers.none_auth_config import (
    ToBroadcasterNoneAuth,
    ToSubscriberNoneAuth,
)
from lonelypsp.auth.helpers.token_auth_config import (
    ToBroadcasterTokenAuth,
    ToSubscriberTokenAuth,
)


def get_auth_config_from_file(
    file_path: str,
) -> Tuple[ToSubscriberAuthConfig, ToBroadcasterAuthConfig]:
    """Reads the incoming/outgoing authorization specified in the file path,
    conventionally called `subscriber-secrets.json`, that was dumped
    from `httppubsubserver --setup`
    """
    with open(file_path, "r") as f:
        raw = json.load(f)

    if raw.get("version") != "2":
        raise ValueError(f"Unknown version {raw['version']}")

    to_subscriber_type = cast(
        Literal["hmac", "token", "none"],
        "none" if "to-subscriber" not in raw else raw["to-subscriber"]["type"],
    )
    to_subscriber_secret = cast(
        Optional[str],
        raw["to-subscriber"]["secret"] if to_subscriber_type != "none" else None,
    )

    to_broadcaster_type = cast(
        Literal["hmac", "token", "none"],
        "none" if "to-broadcaster" not in raw else raw["to-broadcaster"]["type"],
    )
    to_broadcaster_secret = cast(
        Optional[str],
        raw["to-broadcaster"]["secret"] if to_broadcaster_type != "none" else None,
    )

    to_subscriber = cast(Optional[ToSubscriberAuthConfig], None)
    to_broadcaster = cast(Optional[ToBroadcasterAuthConfig], None)

    hmac_db_config: Optional[IncomingHmacAuthDBConfig] = None

    if to_subscriber_type == "none":
        to_subscriber = ToSubscriberNoneAuth()
    elif to_subscriber_type == "token":
        assert to_subscriber_secret is not None, "impossible"
        to_subscriber = ToSubscriberTokenAuth(token=to_subscriber_secret)
    elif to_subscriber_type == "hmac":
        assert to_subscriber_secret is not None, "impossible"
        hmac_db_config = IncomingHmacAuthDBReentrantConfig(
            IncomingHmacAuthSqliteDBConfig(":memory:")
        )
        to_subscriber = ToSubscriberHmacAuth(
            secret=to_subscriber_secret, db_config=hmac_db_config
        )

    if to_broadcaster_type == "none":
        to_broadcaster = ToBroadcasterNoneAuth()
    elif to_broadcaster_type == "token":
        assert to_broadcaster_secret is not None, "impossible"
        to_broadcaster = ToBroadcasterTokenAuth(token=to_broadcaster_secret)
    elif to_broadcaster_type == "hmac":
        assert to_broadcaster_secret is not None, "impossible"
        if hmac_db_config is None:
            hmac_db_config = IncomingHmacAuthDBReentrantConfig(
                IncomingHmacAuthSqliteDBConfig(":memory:")
            )
        to_broadcaster = ToBroadcasterHmacAuth(
            secret=to_broadcaster_secret, db_config=hmac_db_config
        )

    assert (
        to_subscriber is not None
    ), f"unknown or unsupported incoming auth type {to_subscriber_type}"
    assert (
        to_broadcaster is not None
    ), f"unknown or unsupported outgoing auth type {to_broadcaster_type}"

    return (to_subscriber, to_broadcaster)



import os
from typing import List

from lonelypsp.stateful.constants import BroadcasterToSubscriberStatefulMessageType

from lonelypsc.config.config import BroadcastersShuffler
from lonelypsc.util.errors import combine_multiple_exceptions
from lonelypsc.ws.internal_callbacks import (
    inform_internal_message,
)
from lonelypsc.ws.state import (
    ClosingRetryInformationCannotRetry,
    ClosingRetryInformationType,
    ClosingRetryInformationWantRetry,
    InternalMessageStateSent,
    InternalMessageStateType,
    InternalMessageType,
    OpenRetryInformationType,
    ReceivedMessageType,
    ReceivingState,
    RetryInformation,
    SendingState,
    State,
    StateClosing,
    StateOpen,
    StateType,
    TasksOnceOpen,
)


async def recover_open(state: StateOpen, /, *, cause: BaseException) -> State:
    """Recovery function for OPEN; cleans up state-specific resources and
    tries to move towards retrying, raising an error if one occurs during
    cleanup such that recovery is no longer possible
    """
    _cancel_tasks(state)
    bknd_errors = _sweep_backgrounded(state)
    if bknd_errors:
        raise combine_multiple_exceptions("backgrounded tasks failed", bknd_errors)

    tasks_once_open = _build_tasks(state)

    retry: RetryInformation
    if state.retry.type == OpenRetryInformationType.TENTATIVE:
        retry = state.retry.continuation
    else:
        shuffler = BroadcastersShuffler(state.config.broadcasters)
        retry = RetryInformation(
            shuffler=shuffler, iteration=0, iterator=iter(shuffler)
        )

    return StateClosing(
        type=StateType.CLOSING,
        config=state.config,
        cancel_requested=state.cancel_requested,
        broadcaster=state.broadcaster,
        client_session=state.client_session,
        websocket=state.websocket,
        retry=ClosingRetryInformationWantRetry(
            type=ClosingRetryInformationType.WANT_RETRY,
            retry=retry,
            tasks=tasks_once_open,
            exception=cause,
        ),
        backgrounded=state.backgrounded,
    )


def _cancel_tasks(state: StateOpen) -> None:
    """Cancels tasks that are websocket dependent on the state, with
    special handling for the receiving information which may not have
    a single cancelable task for cleanup
    """
    for task in state.compressors.get_compressor_tasks():
        task.cancel()

    if state.receiving is not None:
        try:
            if state.receiving.type == ReceivingState.INCOMPLETE:
                state.receiving.body.close()
                if state.receiving.authorization_task is not None:
                    state.receiving.authorization_task.cancel()
            elif state.receiving.type == ReceivingState.AUTHORIZING:
                state.receiving.body.close()
                state.receiving.authorization_task.cancel()
            elif state.receiving.type == ReceivingState.AUTHORIZING_MISSED:
                state.receiving.authorization_task.cancel()
            elif state.receiving.type == ReceivingState.WAITING_COMPRESSOR:
                state.receiving.compressed_body.close()
            else:
                if not state.receiving.task.cancel():
                    msg = state.receiving.task.result()
                    try:
                        state.received.put_nowait(msg)
                    except BaseException:
                        if msg.type == ReceivedMessageType.LARGE:
                            msg.stream.close()
                        raise
        finally:
            state.receiving = None

    if state.sending is not None:
        state.sending.task.cancel()

    state.read_task.cancel()


def _sweep_backgrounded(state: StateOpen) -> List[BaseException]:
    """Removes done tasks from the states backgrounded set and returns
    any exceptions that were found
    """
    new_backgrounded = set()
    errors: List[BaseException] = []
    for task in state.backgrounded:
        if task.done():
            exc = task.exception()
            if exc is not None:
                errors.append(exc)
        else:
            new_backgrounded.add(task)
    state.backgrounded = new_backgrounded
    return errors


def _build_tasks(state: StateOpen) -> TasksOnceOpen:
    """Builds the tasks that need to be continued with the next connection or
    cleaned up in CLOSING, mutating the state in such a way that this can be
    called multiple times for the same result
    """
    while state.expected_acks:
        ack = state.expected_acks.popleft()
        if (
            ack.type
            == BroadcasterToSubscriberStatefulMessageType.CONFIRM_SUBSCRIBE_EXACT
        ):
            state.exact_subscriptions.add(ack.topic)
        elif (
            ack.type
            == BroadcasterToSubscriberStatefulMessageType.CONFIRM_SUBSCRIBE_GLOB
        ):
            state.glob_subscriptions.add(ack.glob)
        elif (
            ack.type
            == BroadcasterToSubscriberStatefulMessageType.CONFIRM_UNSUBSCRIBE_EXACT
        ):
            state.exact_subscriptions.discard(ack.topic)
        elif (
            ack.type
            == BroadcasterToSubscriberStatefulMessageType.CONFIRM_UNSUBSCRIBE_GLOB
        ):
            state.glob_subscriptions.discard(ack.glob)

    if (
        state.sending is not None
        and state.sending.type == SendingState.INTERNAL_MESSAGE
        and not any(
            m.identifier == state.sending.internal_message.identifier
            for m in state.sent_notifications
        )
    ):
        if state.sending.internal_message.type == InternalMessageType.LARGE:
            state.sending.internal_message.stream.seek(0, os.SEEK_SET)
        state.resending_notifications.append(state.sending.internal_message)
        state.sending = None

    tasks_once_open = TasksOnceOpen(
        exact_subscriptions=state.exact_subscriptions,
        glob_subscriptions=state.glob_subscriptions,
        unsorted=state.management_tasks,
        unsent_notifications=state.unsent_notifications,
        resending_notifications=state.resending_notifications
        + list(state.sent_notifications),
    )

    for internal_message in tasks_once_open.resending_notifications:
        inform_internal_message(
            internal_message,
            InternalMessageStateSent(type=InternalMessageStateType.SENT),
        )

    return tasks_once_open


async def shutdown_open(state: StateOpen, /, *, cause: BaseException) -> StateClosing:
    """Shutdown function for OPEN; cleans up state-specific resources and
    moves towards the CLOSED state. Doesn't raise errors
    """
    _cancel_tasks(state)
    bknd_errors = _sweep_backgrounded(state)
    if bknd_errors:
        cause = combine_multiple_exceptions(
            "irrecoverable error + backgrounded errors", bknd_errors
        )

    tasks_to_cleanup = _build_tasks(state)

    return StateClosing(
        type=StateType.CLOSING,
        config=state.config,
        cancel_requested=state.cancel_requested,
        broadcaster=state.broadcaster,
        client_session=state.client_session,
        websocket=state.websocket,
        retry=ClosingRetryInformationCannotRetry(
            type=ClosingRetryInformationType.CANNOT_RETRY,
            tasks=tasks_to_cleanup,
            exception=cause,
        ),
        backgrounded=state.backgrounded,
    )



from typing import List

from lonelypsc.util.errors import combine_multiple_exceptions
from lonelypsc.ws.handle_connection_failure import (
    cleanup_tasks_and_raise,
    handle_connection_failure,
)
from lonelypsc.ws.internal_callbacks import sweep_internal_message
from lonelypsc.ws.state import (
    ClosingRetryInformationCannotRetry,
    ClosingRetryInformationType,
    State,
    StateType,
)


async def handle_closing(state: State) -> State:
    """Finishes closing the given websocket then either retries with
    the next broadcaster, queues retrying with the next iteration of
    the broadcasters, or gives up
    """
    assert state.type == StateType.CLOSING

    cleanup_excs: List[BaseException] = []
    try:
        await state.websocket.close()
    except BaseException as e:
        cleanup_excs.append(e)

    try:
        await state.client_session.close()
    except BaseException as e:
        cleanup_excs.append(e)

    if state.retry.tasks is not None:
        sweep_excs: List[BaseException] = []
        for msg in state.retry.tasks.resending_notifications:
            try:
                sweep_internal_message(msg)
            except BaseException as e:
                sweep_excs.append(e)

        if sweep_excs:
            if state.retry.type == ClosingRetryInformationType.WANT_RETRY:
                state.retry = ClosingRetryInformationCannotRetry(
                    type=ClosingRetryInformationType.CANNOT_RETRY,
                    tasks=state.retry.tasks,
                    exception=combine_multiple_exceptions(
                        "failed to sweep internal messages before retrying",
                        sweep_excs,
                        context=state.retry.exception,
                    ),
                )
            else:
                cleanup_excs.extend(sweep_excs)

    if (
        cleanup_excs
        and state.retry.type == ClosingRetryInformationType.WANT_RETRY
        and not all(isinstance(e, Exception) for e in cleanup_excs)
    ):
        state.retry = ClosingRetryInformationCannotRetry(
            type=ClosingRetryInformationType.CANNOT_RETRY,
            tasks=state.retry.tasks,
            exception=combine_multiple_exceptions(
                "failed to close websocket before retrying",
                cleanup_excs,
                context=state.retry.exception,
            ),
        )

    if state.retry.type == ClosingRetryInformationType.WANT_RETRY:
        return await handle_connection_failure(
            config=state.config,
            cancel_requested=state.cancel_requested,
            retry=state.retry.retry,
            tasks=state.retry.tasks,
            exception=state.retry.exception,
            backgrounded=state.backgrounded,
        )

    if state.retry.tasks is not None:
        await cleanup_tasks_and_raise(
            state.retry.tasks, state.backgrounded, "closing", state.retry.exception
        )

    raise state.retry.exception



import asyncio
import time
from typing import TYPE_CHECKING

from lonelypsp.stateful.messages.missed import B2S_Missed

from lonelypsc.ws.handlers.open.messages.protocol import MessageChecker
from lonelypsc.ws.handlers.open.websocket_url import (
    make_for_receive_websocket_url_and_change_counter,
)
from lonelypsc.ws.state import (
    ReceivingAuthorizingMissed,
    ReceivingState,
    StateOpen,
)


def check_missed(state: StateOpen, message: B2S_Missed) -> None:
    """Handles the subscriber receiving the a message from the broadcaster that
    the broadcaster may not have sent a message that the subscriber should have
    received
    """
    assert state.receiving is None, "invariant violated: check_missed while receiving"
    state.receiving = ReceivingAuthorizingMissed(
        type=ReceivingState.AUTHORIZING_MISSED,
        message=message,
        authorization_task=asyncio.create_task(
            state.config.is_missed_allowed(
                tracing=message.tracing,
                recovery=make_for_receive_websocket_url_and_change_counter(state),
                topic=message.topic,
                now=time.time(),
                authorization=message.authorization,
            )
        ),
    )


if TYPE_CHECKING:
    _: MessageChecker[B2S_Missed] = check_missed



from typing import TYPE_CHECKING, AsyncIterator, Type

from lonelypsc.util.async_io import AsyncReadableBytesIO


class AsyncIterableAIO:
    """Adapts an AsyncIterable[bytes] to an asynchronous file-like object"""

    def __init__(self, iter: AsyncIterator[bytes]) -> None:
        self._iter: AsyncIterator[bytes] = iter
        self._buffer: bytes = b""
        self._finished: bool = False

    async def read(self, n: int) -> bytes:
        if n < 0:
            while not self._finished:
                try:
                    self._buffer += await self._iter.__anext__()
                except StopAsyncIteration:
                    self._finished = True

            tmp, self._buffer = self._buffer, b""
            return tmp

        while not self._finished and len(self._buffer) < n:
            try:
                self._buffer += await self._iter.__anext__()
            except StopAsyncIteration:
                self._finished = True

        result, self._buffer = self._buffer[:n], self._buffer[n:]
        return result


if TYPE_CHECKING:
    _: Type[AsyncReadableBytesIO] = AsyncIterableAIO



import os
from typing import Union

from lonelypsp.stateful.messages.receive_stream import (
    B2S_ReceiveStreamStartCompressed,
    B2S_ReceiveStreamStartUncompressed,
)
from lonelypsp.stateful.parser_helpers import read_exact

from lonelypsc.types.sync_io import SyncIOBaseLikeIO
from lonelypsc.ws.state import (
    ReceivedLargeMessage,
    ReceivedMessageType,
    ReceivedSmallMessage,
    ReceivingState,
    ReceivingWaitingCompressor,
    StateOpen,
)


def handle_authorized_receive(
    state: StateOpen,
    first: Union[B2S_ReceiveStreamStartUncompressed, B2S_ReceiveStreamStartCompressed],
    body: SyncIOBaseLikeIO,
) -> None:
    """Sets state.receiving as appropriate assuming its either in INCOMPLETE or
    AUTHORIZING state, all the data has been received (and is the right length
    and can be found in body), its authorization task has completed and was ok,
    and the hash of the data matches the expected hash
    """
    spool_size = (
        state.config.max_websocket_message_size
        if state.config.max_websocket_message_size is not None
        else 2**64 - 1
    )

    body.seek(0, os.SEEK_SET)
    if first.compressor_id is None:
        if first.uncompressed_length < spool_size:
            state.received.put_nowait(
                ReceivedSmallMessage(
                    type=ReceivedMessageType.SMALL,
                    topic=first.topic,
                    data=read_exact(body, first.uncompressed_length),
                    sha512=first.unverified_uncompressed_sha512,
                )
            )
            body.close()
            state.receiving = None
            return

        state.received.put_nowait(
            ReceivedLargeMessage(
                type=ReceivedMessageType.LARGE,
                topic=first.topic,
                stream=body,
                sha512=first.unverified_uncompressed_sha512,
            )
        )
        state.receiving = None
        return

    body.seek(0, os.SEEK_SET)
    state.receiving = ReceivingWaitingCompressor(
        type=ReceivingState.WAITING_COMPRESSOR,
        first=first,
        compressed_body=body,
    )



import sys
from typing import List, Optional, Union, cast


def set_context(exc: BaseException, context: Optional[BaseException]) -> None:
    """Sets the context of the exception, merging with the existing one if necessary"""
    if exc.__context__ is None:
        exc.__context__ = context
        return

    if context is None:
        return

    exc.__context__ = combine_multiple_base_exceptions(
        "context", [exc.__context__, context], context=None
    )


if sys.version_info < (3, 11):

    def combine_multiple_base_exceptions(
        msg: str,
        excs: List[BaseException],
        /,
        *,
        context: Optional[BaseException] = None,
    ) -> BaseException:
        """Raises a single BaseException whose __cause__ includes all
        the indicate exceptions and their causes.
        """
        if not excs:
            raise ValueError("no exceptions to combine")

        if len(excs) == 1:
            set_context(excs[0], context)
            return excs[0]

        exc = BaseException(msg)
        last_exc = exc

        for nexc in excs:
            while last_exc.__cause__ is not None:
                last_exc = last_exc.__cause__
            last_exc.__cause__ = nexc

        exc.__context__ = context
        return exc

    def combine_multiple_normal_exceptions(
        msg: str, excs: List[Exception], /, *, context: Optional[BaseException] = None
    ) -> Exception:
        """Raises a single Exception whose __cause__ includes all
        the indicate exceptions and their causes.
        """
        if not excs:
            raise ValueError("no exceptions to combine")

        if len(excs) == 1:
            set_context(excs[0], context)
            return excs[0]

        exc = Exception(msg)
        last_exc: Union[Exception, BaseException] = exc

        for nexc in excs:
            while last_exc.__cause__ is not None:
                last_exc = last_exc.__cause__
            last_exc.__cause__ = nexc

        exc.__context__ = context
        return exc

else:

    def combine_multiple_base_exceptions(
        msg: str,
        excs: List[BaseException],
        /,
        *,
        context: Optional[BaseException] = None,
    ) -> BaseException:
        """Light wrapper around BaseExceptionGroup"""
        if not excs:
            raise ValueError("no exceptions to combine")

        if len(excs) == 1:
            set_context(excs[0], context)
            return excs[0]

        if any(isinstance(e, BaseExceptionGroup) for e in excs):
            new_excs: List[BaseException] = []
            for e in excs:
                if isinstance(e, BaseExceptionGroup):
                    new_excs.extend(e.exceptions)
                else:
                    new_excs.append(e)
            excs = new_excs

        result = BaseExceptionGroup(msg, excs)
        result.__context__ = context
        return result

    def combine_multiple_normal_exceptions(
        msg: str, excs: List[Exception], /, *, context: Optional[BaseException] = None
    ) -> Exception:
        """Light wrapper around ExceptionGroup"""
        if not excs:
            raise ValueError("no exceptions to combine")

        if len(excs) == 1:
            set_context(excs[0], context)
            return excs[0]

        if any(isinstance(e, ExceptionGroup) for e in excs):
            new_excs: List[Exception] = []
            for e in excs:
                if isinstance(e, ExceptionGroup):
                    new_excs.extend(e.exceptions)
                else:
                    new_excs.append(e)
            excs = new_excs

        result = ExceptionGroup(msg, excs)
        result.__context__ = context
        return result


def combine_multiple_exceptions(
    msg: str, excs: List[BaseException], /, *, context: Optional[BaseException] = None
) -> BaseException:
    """Returns a single Exception, if possible, otherwise a BaseException which will
    report all the indicated exceptions in the most informative way possible.
    """
    if all(isinstance(e, Exception) for e in excs):
        return combine_multiple_normal_exceptions(
            msg, cast(List[Exception], excs), context=context
        )
    return combine_multiple_base_exceptions(msg, excs, context=context)



import asyncio
import base64
import hashlib
import io
import random
import tempfile
import time
from typing import (
    TYPE_CHECKING,
    Annotated,
    ContextManager,
    Dict,
    Generic,
    List,
    Literal,
    Optional,
    Protocol,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
)

import aiohttp
from aiohttp.typedefs import LooseHeaders
from fastapi import APIRouter, Header
from fastapi.requests import Request
from fastapi.responses import Response
from lonelypsp.auth.config import AuthResult
from lonelypsp.compat import assert_never
from lonelypsp.stateful.parser_helpers import read_exact
from lonelypsp.stateless.constants import (
    BroadcasterToSubscriberStatelessMessageType,
    SubscriberToBroadcasterStatelessMessageType,
)
from lonelypsp.stateless.make_strong_etag import (
    GlobAndRecovery,
    StrongEtag,
    TopicAndRecovery,
    make_strong_etag,
)
from lonelypsp.tracing.stateless.notify import (
    StatelessTracingNotifyOnSending,
)
from lonelypsp.util.cancel_and_check import cancel_and_check
from starlette.background import BackgroundTask

from lonelypsc.client import (
    PubSubClient,
    PubSubClientBulkSubscriptionConnector,
    PubSubClientConnectionStatus,
    PubSubClientConnector,
    PubSubClientMessageWithCleanup,
    PubSubClientReceiver,
    PubSubClientTracingNotifyStart,
    PubSubDirectConnectionStatusReceiver,
    PubSubDirectOnMessageWithCleanupReceiver,
    PubSubRequestAmbiguousError,
    PubSubRequestError,
    PubSubRequestRefusedError,
    PubSubRequestRetriesExhaustedError,
)
from lonelypsc.config.config import BroadcastersShuffler, PubSubBroadcasterConfig
from lonelypsc.config.helpers.uvicorn_bind_config import handle_bind_with_uvicorn
from lonelypsc.config.http_config import HttpPubSubConfig
from lonelypsc.http.notify.action import HttpNotifyAction, HttpNotifyInfo
from lonelypsc.http.retrier import (
    BroadcasterRetryableActionResult,
    attempt_broadcasters,
)
from lonelypsc.types.sync_io import SyncStandardIO
from lonelypsc.util.async_io import async_read_exact
from lonelypsc.util.errors import combine_multiple_exceptions
from lonelypsc.util.io_helpers import (
    PositionedSyncStandardIO,
)
from lonelypsc.util.request_body_io import AsyncIterableAIO

if TYPE_CHECKING:
    from lonelypsc.http.notify.result import HttpPubSubNotifyResult

T = TypeVar("T")
T_co = TypeVar("T_co", covariant=True)
InitializerT = TypeVar("InitializerT")


class _BroadcasterCallable(Protocol[T_co]):
    async def __call__(self, /, *, broadcaster: PubSubBroadcasterConfig) -> T_co:
        raise NotImplementedError


class HttpPubSubClientConnector(Generic[InitializerT]):
    def __init__(self, config: HttpPubSubConfig[InitializerT]) -> None:
        self.config = config
        """The configuration that dictates how we behave"""

        self._session: Optional[aiohttp.ClientSession] = None
        """The client session to use for making requests, if entered, otherwise None"""

        self._shuffler: Optional[BroadcastersShuffler] = None
        """The shuffler for the broadcasters list, if entered, otherwise None"""

    async def setup_connector(self) -> None:
        assert self._session is None, "already set up"
        sess = aiohttp.ClientSession(
            timeout=aiohttp.ClientTimeout(
                total=self.config.outgoing_http_timeout_total,
                connect=self.config.outgoing_http_timeout_connect,
                sock_read=self.config.outgoing_http_timeout_sock_read,
                sock_connect=self.config.outgoing_http_timeout_sock_connect,
            ),
            skip_auto_headers=("User-Agent",),
            auto_decompress=False,
        )
        await sess.__aenter__()
        self._session = sess
        self._shuffler = BroadcastersShuffler(self.config.broadcasters)

    async def teardown_connector(self) -> None:
        assert self._session is not None, "not set up"
        sess = self._session
        self._session = None
        self._shuffler = None
        await sess.__aexit__(None, None, None)
        return None

    async def _try_large_post_request(
        self,
        /,
        *,
        broadcaster: PubSubBroadcasterConfig,
        headers: LooseHeaders,
        path: str,
        body: SyncStandardIO,
        body_starts_at: int,
    ) -> Union[
        aiohttp.ClientResponse,
        Literal["ambiguous", "refused", "retry"],
    ]:
        """Not async, thread, or process safe when reusing body.

        - MUST specify content-length
        - MUST specify content-type
        """
        assert self._session is not None, "not set up"

        body.seek(body_starts_at)
        try:
            result = await self._session.post(
                broadcaster["host"] + path,
                data=body,
                headers=headers,
                allow_redirects=False,
                read_until_eof=False,
            )
        except aiohttp.ClientError:
            return "ambiguous"

        await result.__aenter__()
        if result.status in (502, 503, 504):
            await result.__aexit__(None, None, None)
            return "retry"

        if result.status < 200 or result.status >= 300:
            await result.__aexit__(None, None, None)
            return "refused"

        return result

    async def _retry_with_broadcasters(
        self, /, *, broadcaster_callable: _BroadcasterCallable[T]
    ) -> T:
        """Calls the given function with each broadcaster in a random order up to the
        configured number of retries. async safe iff broadcaster_callable is async safe.

        `T` should include `Literal["retry", "refused", "ambiguous"]` to indicate that the
        request was not received by the broadcaster and should be retried, and anything
        else to indicate success.
        """
        assert self._session is not None, "not set up"
        assert self._shuffler is not None, "not set up"

        # this could be a boolean, but doing it this way helps static
        # analysis understand we don't return ambiguous unless T is a superset
        # of Literal["ambiguous"]
        seen_ambiguous: Optional[T] = None

        for iteration in range(self.config.outgoing_retries_per_broadcaster):
            if iteration > 0:
                await asyncio.sleep(2 ** (iteration - 1) + random.random())

            for broadcaster in self._shuffler:
                result = await broadcaster_callable(broadcaster=broadcaster)
                if result != "ambiguous" and result != "retry":
                    return result
                if result == "ambiguous" and not self.config.outgoing_retry_ambiguous:
                    return result
                if result == "ambiguous":
                    seen_ambiguous = result

        if seen_ambiguous is not None:
            return seen_ambiguous
        return result

    async def _make_large_post_request(
        self, /, *, path: str, headers: LooseHeaders, body: SyncStandardIO
    ) -> Union[aiohttp.ClientResponse, Literal["retry", "refused", "ambiguous"]]:
        """Not async, thread, or process safe when reusing body. Assumes the body and/or
        response may be very large; if the request body is large, it should already be
        spooled if necessary and this will rewind when required

        Returned client response is already entered but not released

        - MUST specify content-length
        - MUST specify content-type

        Result is a response which may not indicate success, but should definitely not
        be retried
        """
        assert self._session is not None, "not set up"

        body_starts_at = body.tell()

        async def broadcaster_callable(
            *,
            broadcaster: PubSubBroadcasterConfig,
        ) -> Union[aiohttp.ClientResponse, Literal["retry", "refused", "ambiguous"]]:
            return await self._try_large_post_request(
                broadcaster=broadcaster,
                headers=headers,
                path=path,
                body=body,
                body_starts_at=body_starts_at,
            )

        return await self._retry_with_broadcasters(
            broadcaster_callable=broadcaster_callable
        )

    async def _try_small_request(
        self,
        /,
        *,
        method: Literal["GET", "POST"],
        broadcaster: PubSubBroadcasterConfig,
        headers: LooseHeaders,
        path: str,
        body: Optional[bytes],
        special_ok_codes: Set[int],
    ) -> Union[bytes, Literal["ambiguous", "retry", "refused"]]:
        """Tries the given broadcaster with a post/get request, assuming everything
        can be held in memory. async safe.
        """
        assert self._session is not None, "not set up"

        try:
            async with self._session.request(
                method,
                broadcaster["host"] + path,
                headers=headers,
                data=body,
                allow_redirects=False,
            ) as resp:
                if resp.status not in special_ok_codes:
                    if resp.status in (502, 503, 504):
                        return "retry"
                    if resp.status < 200 or resp.status >= 300:
                        return "refused"
                return await resp.read()
        except aiohttp.ClientError:
            return "ambiguous"

    async def _make_small_request(
        self,
        /,
        *,
        method: Literal["GET", "POST"],
        headers: LooseHeaders,
        path: str,
        body: Optional[bytes],
        special_ok_codes: Set[int],
    ) -> Union[bytes, Literal["ambiguous", "refused", "retry"]]:
        """Makes a small request, trying broadcasters in a random order up to
        the configured number of retries. async safe.
        """
        assert self._session is not None, "not set up"

        async def broadcaster_callable(
            *,
            broadcaster: PubSubBroadcasterConfig,
        ) -> Union[Literal["ambiguous", "retry", "refused"], bytes]:
            return await self._try_small_request(
                method=method,
                broadcaster=broadcaster,
                headers=headers,
                path=path,
                body=body,
                special_ok_codes=special_ok_codes,
            )

        return await self._retry_with_broadcasters(
            broadcaster_callable=broadcaster_callable
        )

    @property
    def _receive_url(self) -> str:
        host_url = self.config.host
        host_fragment_starts_at = host_url.find("#")
        host_fragment = ""
        if host_fragment_starts_at != -1:
            host_fragment = host_url[host_fragment_starts_at:]
            host_url = host_url[:host_fragment_starts_at]

        return host_url + "/v1/receive" + host_fragment

    @property
    def _recovery_url(self) -> str:
        host_url = self.config.host
        host_fragment_starts_at = host_url.find("#")
        host_fragment = ""
        if host_fragment_starts_at != -1:
            host_fragment = host_url[host_fragment_starts_at:]
            host_url = host_url[:host_fragment_starts_at]

        return host_url + "/v1/recover" + host_fragment

    def _raise_for_error(
        self, /, result: Union[Literal["ambiguous", "retry", "refused"], T]
    ) -> None:
        if result == "ambiguous":
            raise PubSubRequestAmbiguousError()
        if result == "retry":
            raise PubSubRequestRetriesExhaustedError()
        if result == "refused":
            raise PubSubRequestRefusedError()

    async def subscribe_exact(self, /, *, topic: bytes) -> None:
        assert self._session is not None, "not set up"
        receive_url = self._receive_url
        recovery_url = self._recovery_url
        tracing = b""  # TODO: tracing

        auth_at = time.time()
        authorization = await self.config.authorize_subscribe_exact(
            tracing=tracing,
            url=receive_url,
            recovery=recovery_url,
            exact=topic,
            now=auth_at,
        )
        headers: Dict[str, str] = {
            "Content-Type": "application/octet-stream",
        }
        if authorization is not None:
            headers["Authorization"] = authorization

        encoded_receive_url = receive_url.encode("utf-8")
        encoded_recovery_url = recovery_url.encode("utf-8")

        body = io.BytesIO()
        body.write(len(encoded_receive_url).to_bytes(2, "big", signed=False))
        body.write(encoded_receive_url)
        body.write(len(topic).to_bytes(2, "big", signed=False))
        body.write(topic)
        body.write(len(encoded_recovery_url).to_bytes(2, "big", signed=False))
        body.write(encoded_recovery_url)
        body.write(len(tracing).to_bytes(2, "big", signed=False))
        body.write(tracing)

        result = await self._make_small_request(
            method="POST",
            headers=headers,
            path="/v1/subscribe/exact",
            body=body.getvalue(),
            special_ok_codes={
                409
            },  # TODO: treat as failure (auth is being skipped) and update docs
        )
        self._raise_for_error(result)
        # TODO: check response auth

    async def subscribe_glob(self, /, *, glob: str) -> None:
        assert self._session is not None, "not set up"
        receive_url = self._receive_url
        recovery_url = self._recovery_url
        tracing = b""

        auth_at = time.time()
        authorization = await self.config.authorize_subscribe_glob(
            tracing=tracing,
            url=receive_url,
            recovery=recovery_url,
            glob=glob,
            now=auth_at,
        )
        headers: Dict[str, str] = {
            "Content-Type": "application/octet-stream",
        }
        if authorization is not None:
            headers["Authorization"] = authorization

        encoded_receive_url = receive_url.encode("utf-8")
        encoded_glob = glob.encode("utf-8")
        encoded_recovery_url = recovery_url.encode("utf-8")

        body = io.BytesIO()
        body.write(len(encoded_receive_url).to_bytes(2, "big", signed=False))
        body.write(encoded_receive_url)
        body.write(len(encoded_glob).to_bytes(2, "big", signed=False))
        body.write(encoded_glob)
        body.write(len(encoded_recovery_url).to_bytes(2, "big", signed=False))
        body.write(encoded_recovery_url)
        body.write(len(tracing).to_bytes(2, "big", signed=False))
        body.write(tracing)

        result = await self._make_small_request(
            method="POST",
            headers=headers,
            path="/v1/subscribe/glob",
            body=body.getvalue(),
            special_ok_codes={409},  # TODO: treat as failure and update docs
        )
        self._raise_for_error(result)
        # TODO: check response auth

    async def unsubscribe_exact(self, /, *, topic: bytes) -> None:
        assert self._session is not None, "not set up"
        receive_url = self._receive_url

        auth_at = time.time()
        tracing = b""  # TODO: tracing
        authorization = await self.config.authorize_subscribe_exact(
            tracing=tracing, url=receive_url, recovery=None, exact=topic, now=auth_at
        )
        headers: Dict[str, str] = {
            "Content-Type": "application/octet-stream",
        }
        if authorization is not None:
            headers["Authorization"] = authorization

        encoded_receive_url = receive_url.encode("utf-8")

        body = io.BytesIO()
        body.write(len(encoded_receive_url).to_bytes(2, "big", signed=False))
        body.write(encoded_receive_url)
        body.write(len(topic).to_bytes(2, "big", signed=False))
        body.write(topic)
        body.write(len(tracing).to_bytes(2, "big", signed=False))
        body.write(tracing)

        result = await self._make_small_request(
            method="POST",
            headers=headers,
            path="/v1/unsubscribe/exact",
            body=body.getvalue(),
            special_ok_codes={409},  # TODO: treat as failure and update docs
        )
        self._raise_for_error(result)

        # TODO: check response auth

    async def unsubscribe_glob(self, /, *, glob: str) -> None:
        assert self._session is not None, "not set up"
        receive_url = self._receive_url

        tracing = b""  # TODO: traicng
        auth_at = time.time()
        authorization = await self.config.authorize_subscribe_glob(
            tracing=tracing, url=receive_url, recovery=None, glob=glob, now=auth_at
        )
        headers: Dict[str, str] = {
            "Content-Type": "application/octet-stream",
        }
        if authorization is not None:
            headers["Authorization"] = authorization

        encoded_receive_url = receive_url.encode("utf-8")
        encoded_glob = glob.encode("utf-8")

        body = io.BytesIO()
        body.write(len(encoded_receive_url).to_bytes(2, "big", signed=False))
        body.write(encoded_receive_url)
        body.write(len(encoded_glob).to_bytes(2, "big", signed=False))
        body.write(encoded_glob)
        body.write(len(tracing).to_bytes(2, "big", signed=False))
        body.write(tracing)

        result = await self._make_small_request(
            method="POST",
            headers=headers,
            path="/v1/unsubscribe/glob",
            body=body.getvalue(),
            special_ok_codes={409},  # TODO: treat as failure and update docs
        )
        self._raise_for_error(result)
        # TODO: check response auth

    def get_bulk(self) -> Optional[PubSubClientBulkSubscriptionConnector]:
        return self

    async def check_subscriptions(self) -> StrongEtag:
        assert self._session is not None, "not set up"
        receive_url = self._receive_url

        tracing = b""  # TODO: tracing
        auth_at = time.time()
        authorization = await self.config.authorize_check_subscriptions(
            tracing=tracing, url=receive_url, now=auth_at
        )
        headers: Dict[str, str] = {
            "Content-Type": "application/octet-stream",
        }
        if authorization is not None:
            headers["Authorization"] = authorization

        encoded_receive_url = receive_url.encode("utf-8")

        body = io.BytesIO()
        body.write(len(tracing).to_bytes(2, "big", signed=False))
        body.write(tracing)
        body.write(len(encoded_receive_url).to_bytes(2, "big", signed=False))
        body.write(encoded_receive_url)

        result = await self._make_small_request(
            method="POST",
            headers=headers,
            path="/v1/check_subscriptions",
            body=body.getvalue(),
            special_ok_codes=set(),
        )
        self._raise_for_error(result)
        assert isinstance(result, bytes), "impossible"

        rdr = io.BytesIO(result)
        response_type_bytes = read_exact(rdr, 2)
        response_type = int.from_bytes(response_type_bytes, "big", signed=False)

        assert response_type == int(
            BroadcasterToSubscriberStatelessMessageType.RESPONSE_CHECK_SUBSCRIPTIONS
        ), f"{response_type=}"

        resp_authorization_length_bytes = rdr.read(2)
        resp_authorization_length = int.from_bytes(
            resp_authorization_length_bytes, "big", signed=False
        )
        resp_authorization_bytes = read_exact(rdr, resp_authorization_length)
        resp_authorization = (
            None
            if resp_authorization_bytes == b""
            else resp_authorization_bytes.decode("utf-8")
        )

        resp_tracing_length_bytes = rdr.read(2)
        resp_tracing_length = int.from_bytes(
            resp_tracing_length_bytes, "big", signed=False
        )
        resp_tracing = read_exact(rdr, resp_tracing_length)

        etag_format_bytes = rdr.read(1)
        etag_format = int.from_bytes(etag_format_bytes, "big", signed=False)

        if etag_format != 0:
            raise ValueError("invalid strong etag format")

        etag = rdr.read(64)
        strong_etag = StrongEtag(format=0, etag=etag)

        if not self.config.is_check_subscription_response_allowed(
            tracing=resp_tracing,
            strong_etag=strong_etag,
            authorization=resp_authorization,
            now=time.time(),
        ):
            raise PubSubRequestError(
                "check subscription response failed authorization check"
            )

        return strong_etag

    async def set_subscriptions(
        self,
        /,
        *,
        exact: List[bytes],
        globs: List[str],
    ) -> None:
        assert self._session is not None, "not set up"
        recovery_url = self._recovery_url
        encoded_recovery_url = recovery_url.encode("utf-8")

        rec_and_length = bytearray(2 + len(encoded_recovery_url))
        rec_and_length[:2] = len(encoded_recovery_url).to_bytes(2, "big", signed=False)
        rec_and_length[2:] = encoded_recovery_url

        exact = sorted(exact)
        globs = sorted(globs)

        receive_url = self._receive_url

        strong_etag = make_strong_etag(
            receive_url,
            [TopicAndRecovery(t, recovery_url) for t in exact],
            [GlobAndRecovery(g, recovery_url) for g in globs],
            recheck_sort=False,
        )

        tracing = b""  # TODO: tracing
        auth_at = time.time()
        authorization = await self.config.authorize_set_subscriptions(
            tracing=tracing, url=receive_url, strong_etag=strong_etag, now=auth_at
        )
        headers: Dict[str, str] = {
            "Content-Type": "application/octet-stream",
        }
        if authorization is not None:
            headers["Authorization"] = authorization

        encoded_receive_url = receive_url.encode("utf-8")
        body = io.BytesIO()
        body.write(len(tracing).to_bytes(2, "big", signed=False))
        body.write(tracing)
        body.write(len(encoded_receive_url).to_bytes(2, "big", signed=False))
        body.write(encoded_receive_url)
        body.write(strong_etag.format.to_bytes(1, "big", signed=False))
        body.write(strong_etag.etag)
        body.write(len(exact).to_bytes(4, "big", signed=False))
        for topic in exact:
            body.write(len(topic).to_bytes(2, "big", signed=False))
            body.write(topic)
            body.write(rec_and_length)

        body.write(len(globs).to_bytes(4, "big", signed=False))
        for glob in globs:
            encoded_glob = glob.encode("utf-8")
            body.write(len(encoded_glob).to_bytes(2, "big", signed=False))
            body.write(encoded_glob)
            body.write(rec_and_length)

        result = await self._make_small_request(
            method="POST",
            headers=headers,
            path="/v1/set_subscriptions",
            body=body.getvalue(),
            special_ok_codes=set(),
        )
        self._raise_for_error(result)

        # TODO: check response auth

    def prepare_notifier_trace(
        self, initializer: InitializerT, /
    ) -> ContextManager[
        PubSubClientTracingNotifyStart[StatelessTracingNotifyOnSending]
    ]:
        return self.config.tracing.notify(initializer)

    async def notify(
        self,
        /,
        *,
        topic: bytes,
        message: SyncStandardIO,
        length: int,
        message_sha512: bytes,
        tracer: StatelessTracingNotifyOnSending,
    ) -> HttpPubSubNotifyResult:
        assert self._session is not None, "not set up"
        assert self._shuffler is not None, "not set up"
        initial_message_tell = message.tell()
        normalized_message = PositionedSyncStandardIO(
            message,
            start_idx=initial_message_tell,
            end_idx=initial_message_tell + length,
        )
        result = await attempt_broadcasters(
            config=self.config,
            shuffler=self._shuffler,
            action=HttpNotifyAction(
                config=self.config,
                info=HttpNotifyInfo(
                    topic=topic,
                    normalized_message=normalized_message,
                    sha512=message_sha512,
                ),
                session=self._session,
                tracer=tracer,
                seen_ambiguous=False,
            ),
        )
        if result.type == BroadcasterRetryableActionResult.FAILURE:
            self._raise_for_error(result.result)
            assert False, "unreachable"
        if result.type != BroadcasterRetryableActionResult.SUCCESS:
            assert_never(result)
        return result.result


if TYPE_CHECKING:
    __: Type[PubSubClientConnector] = HttpPubSubClientConnector


class HttpPubSubClientReceiver:
    def __init__(self, config: HttpPubSubConfig) -> None:
        self.config = config
        self.handlers: List[Tuple[int, PubSubDirectOnMessageWithCleanupReceiver]] = []
        """The registered on_message receivers"""
        self.status_handlers: List[Tuple[int, PubSubDirectConnectionStatusReceiver]] = (
            []
        )
        """The registered connection status handlers"""
        self.bind_task: Optional[asyncio.Task] = None
        self.connection_status = PubSubClientConnectionStatus.LOST
        self._status_counter = 0
        """Ensures we can give unique status handler ids"""
        self._status_handlers_lock = asyncio.Lock()
        """lock for interacting with status_handlers"""

    async def setup_receiver(self) -> None:
        assert self.bind_task is None, "already setup & not re-entrant"
        assert (
            self.connection_status == PubSubClientConnectionStatus.LOST
        ), "previously setup and not reusable"
        bind_config = self.config.bind

        if bind_config["type"] == "uvicorn":
            bind_config = await handle_bind_with_uvicorn(bind_config)

        router = APIRouter()
        router.add_api_route("/v1/receive", self._receive, methods=["POST"])
        router.add_api_route("/v1/recover", self._recover, methods=["POST"])
        self.bind_task = asyncio.create_task(bind_config["callback"](router))

        async with self._status_handlers_lock:
            for _, status_handler in self.status_handlers:
                try:
                    await status_handler.on_connection_established()
                except BaseException as e:
                    excs: List[BaseException] = [e]
                    self.connection_status = PubSubClientConnectionStatus.ABANDONED

                    canceler = asyncio.create_task(cancel_and_check(self.bind_task))
                    self.bind_task = None

                    for _, handler in self.status_handlers:
                        try:
                            await handler.on_connection_abandoned()
                        except BaseException as e2:
                            excs.append(e2)

                    self.status_handlers = []
                    try:
                        await canceler
                    except BaseException as e:
                        excs.append(e)

                    raise combine_multiple_exceptions(
                        "failed to tell status handlers connection established", excs
                    )

            self.connection_status = PubSubClientConnectionStatus.OK

    async def teardown_receiver(self) -> None:
        assert self.bind_task is not None, "not set up"
        canceler = asyncio.create_task(cancel_and_check(self.bind_task))
        self.bind_task = None

        excs: List[BaseException] = []
        async with self._status_handlers_lock:
            self.connection_status = PubSubClientConnectionStatus.ABANDONED
            for _, handler in self.status_handlers:
                try:
                    await handler.on_connection_abandoned()
                except BaseException as e:
                    excs.append(e)

            self.status_handlers = []

        try:
            await canceler
        except BaseException as e:
            excs.append(e)

        if excs:
            raise combine_multiple_exceptions("failed to teardown receiver", excs)

    async def _receive(
        self,
        request: Request,
        authorization: Annotated[Optional[str], Header()] = None,
        repr_digest: Annotated[Optional[str], Header()] = None,
    ) -> Response:
        """HttpPubSubClientReceiver primary endpoint

        The authorization header provided shows that the request came from a broadcaster,
        and is validated according to the `auth` mechanism configured.

        The `Repr-Digest` header MUST include the sha-512 digest of the message. The repr
        digest is used to bail out early if the request is not authorized, but is rechecked
        before processing. It MAY include additional digests in any order.

        The `X-Topic` header MUST be set to the topic name, base64 encoded.
        """
        if self.connection_status == PubSubClientConnectionStatus.ABANDONED:
            return Response(
                status_code=503,
                background=BackgroundTask(
                    _raise_excs,
                    "connection status abandoned but received message",
                    [Exception("receive after abandoned")],
                ),
            )

        if repr_digest is None:
            return Response(
                status_code=400,
                headers={"Content-Type": "application/json; charset=utf-8"},
                content=b'{"unsubscribe": true, "reason": "missing repr-digest header"}',
            )

        expected_digest_b64: Optional[str] = None
        for digest_pair in repr_digest.split(","):
            split_digest_pair = digest_pair.split("=", 1)
            if len(split_digest_pair) != 2:
                continue
            digest_type, digest_value = split_digest_pair
            if digest_type != "sha-512":
                continue

            expected_digest_b64 = digest_value

        if expected_digest_b64 is None:
            return Response(
                status_code=400,
                headers={"Content-Type": "application/json; charset=utf-8"},
                content=b'{"unsubscribe": true, "reason": "missing sha-512 repr-digest"}',
            )

        try:
            expected_digest = base64.b64decode(expected_digest_b64 + "==")
        except BaseException:
            return Response(
                status_code=400,
                headers={"Content-Type": "application/json; charset=utf-8"},
                content=b'{"unsubscribe": true, "reason": "unparseable sha-512 repr-digest (not base64)"}',
            )

        stream = request.stream()
        try:
            body = AsyncIterableAIO(stream)

            try:
                tracing_length_bytes = await async_read_exact(body, 2)
                tracing_length = int.from_bytes(tracing_length_bytes, "big")
                tracing = await async_read_exact(body, tracing_length)

                topic_length_bytes = await async_read_exact(body, 2)
                topic_length = int.from_bytes(topic_length_bytes, "big")
                topic = await async_read_exact(body, topic_length)

                identifier_length_bytes = await async_read_exact(body, 1)
                identifier_length = int.from_bytes(identifier_length_bytes, "big")
                identifier = await async_read_exact(body, identifier_length)

                message_length_bytes = await async_read_exact(body, 8)
                message_length = int.from_bytes(message_length_bytes, "big")
            except ValueError:
                return Response(status_code=400)

            auth_result = await self.config.is_receive_allowed(
                tracing=tracing,
                url=str(request.url),
                topic=topic,
                message_sha512=expected_digest,
                identifier=identifier,
                now=time.time(),
                authorization=authorization,
            )
            if auth_result == AuthResult.UNAVAILABLE:
                async with self._status_handlers_lock:
                    if self.connection_status == PubSubClientConnectionStatus.OK:
                        self.connection_status = PubSubClientConnectionStatus.LOST
                        excs: List[BaseException] = []
                        for _, status_handler in self.status_handlers:
                            try:
                                await status_handler.on_connection_lost()
                            except BaseException as e:
                                excs.append(e)

                return Response(
                    status_code=503,
                    background=(
                        BackgroundTask(
                            _raise_excs,
                            "db auth unavailable then failed to inform status handlers",
                            excs,
                        )
                        if excs
                        else None
                    ),
                )

            if auth_result != AuthResult.OK:
                return Response(
                    status_code=403,
                    headers={"Content-Type": "application/octet-stream"},
                    content=int(
                        SubscriberToBroadcasterStatelessMessageType.RESPONSE_UNSUBSCRIBE_IMMEDIATE
                    ).to_bytes(2, "big"),
                )

            background_exceptions: List[BaseException] = []
            async with self._status_handlers_lock:
                if self.connection_status == PubSubClientConnectionStatus.LOST:
                    self.connection_status = PubSubClientConnectionStatus.OK
                    for _, status_handler in self.status_handlers:
                        try:
                            await status_handler.on_connection_established()
                        except BaseException as e:
                            background_exceptions.append(e)

            with tempfile.SpooledTemporaryFile(
                max_size=self.config.message_body_spool_size, mode="w+b"
            ) as spooled_request_body:
                read_length = 0
                hasher = hashlib.sha512()
                while True:
                    chunk = await body.read(io.DEFAULT_BUFFER_SIZE)
                    if not chunk:
                        break
                    hasher.update(chunk)
                    read_length += len(chunk)
                    spooled_request_body.write(chunk)

                    if read_length > message_length:
                        break

                if read_length != message_length:
                    background_exceptions.append(Exception("incorrect message length"))
                    return Response(
                        status_code=400,
                        headers={"Content-Type": "application/octet-stream"},
                        content=int(
                            SubscriberToBroadcasterStatelessMessageType.RESPONSE_UNSUBSCRIBE_IMMEDIATE
                        ).to_bytes(2, "big"),
                        background=(
                            BackgroundTask(
                                _raise_excs,
                                "incorrect message length",
                                background_exceptions,
                            )
                        ),
                    )

                real_digest = hasher.digest()
                if real_digest != expected_digest:
                    background_exceptions.append(
                        Exception("incorrect sha-512 repr-digest")
                    )
                    return Response(
                        status_code=403,
                        headers={"Content-Type": "application/octet-stream"},
                        content=int(
                            SubscriberToBroadcasterStatelessMessageType.RESPONSE_UNSUBSCRIBE_IMMEDIATE
                        ).to_bytes(2, "big"),
                        background=(
                            BackgroundTask(
                                _raise_excs,
                                "incorrect sha-512 repr-digest",
                                background_exceptions,
                            )
                        ),
                    )

                for idx, (reg_id, handler) in enumerate(tuple(self.handlers)):
                    if len(self.handlers) <= idx or self.handlers[idx][0] != reg_id:
                        for alt_idx in range(min(idx, len(self.handlers))):
                            if self.handlers[alt_idx][0] == reg_id:
                                break
                        else:
                            continue
                    spooled_request_body.seek(0)

                    # we want message.cleanup() not to interfere with future
                    # handlers if called multiple times, hence we make a new event
                    # per handler
                    handler_done = asyncio.Event()

                    async def handler_cleanup() -> None:
                        handler_done.set()

                    message = PubSubClientMessageWithCleanup(
                        topic=topic,
                        sha512=real_digest,
                        data=spooled_request_body,
                        cleanup=handler_cleanup,
                    )
                    try:
                        await handler.on_message(message)
                        await handler_done.wait()
                    except asyncio.CancelledError as canceled:
                        background_exceptions.append(canceled)
                        raise combine_multiple_exceptions(
                            "handler canceled", background_exceptions
                        )
                    except BaseException as e:
                        background_exceptions.append(e)
        finally:
            await stream.aclose()

        resp_tracing = b""  # TODO: tracing
        num_subscribers = 1
        resp_authorization = await self.config.authorize_confirm_receive(
            tracing=resp_tracing,
            identifier=identifier,
            num_subscribers=num_subscribers,
            url=str(request.url),
            now=time.time(),
        )
        resp_authorization_bytes = (
            b"" if resp_authorization is None else resp_authorization.encode("utf-8")
        )

        response_body = io.BytesIO()
        response_body.write(
            int(
                SubscriberToBroadcasterStatelessMessageType.RESPONSE_CONFIRM_RECEIVE
            ).to_bytes(2, "big")
        )
        response_body.write(len(resp_authorization_bytes).to_bytes(2, "big"))
        response_body.write(resp_authorization_bytes)
        response_body.write(len(resp_tracing).to_bytes(2, "big"))
        response_body.write(resp_tracing)
        response_body.write(len(identifier).to_bytes(1, "big"))
        response_body.write(identifier)
        response_body.write(num_subscribers.to_bytes(4, "big"))
        return Response(
            status_code=200,
            headers={"Content-Type": "application/octet-stream"},
            content=response_body.getvalue(),
            background=(
                BackgroundTask(
                    _raise_excs,
                    "failed to inform handlers about received message",
                    background_exceptions,
                )
                if background_exceptions
                else None
            ),
        )

    async def _recover(
        self,
        request: Request,
        authorization: Annotated[Optional[str], Header()] = None,
    ) -> Response:
        stream = request.stream()
        try:
            body = AsyncIterableAIO(stream.__aiter__())
            tracing_length_bytes = await async_read_exact(body, 2)
            tracing_length = int.from_bytes(tracing_length_bytes, "big")
            tracing = await async_read_exact(body, tracing_length)
            topic_length_bytes = await async_read_exact(body, 2)
            topic_length = int.from_bytes(topic_length_bytes, "big")
            topic = await async_read_exact(body, topic_length)
        finally:
            await stream.aclose()

        auth_result = await self.config.is_missed_allowed(
            tracing=tracing,
            recovery=str(request.url),
            topic=topic,
            now=time.time(),
            authorization=authorization,
        )
        background_exceptions: List[BaseException] = []
        if auth_result == AuthResult.UNAVAILABLE:
            background_exceptions.append(Exception("auth unavailable"))
            return Response(
                status_code=503,
                background=BackgroundTask(
                    _raise_excs, "auth unavailable", background_exceptions
                ),
            )
        if auth_result != AuthResult.OK:
            background_exceptions.append(Exception(f"auth failed: {auth_result=}"))
            return Response(
                status_code=403,
                background=BackgroundTask(
                    _raise_excs, "auth failed", background_exceptions
                ),
            )

        async with self._status_handlers_lock:
            if self.connection_status == PubSubClientConnectionStatus.ABANDONED:
                return Response(status_code=503)

            if self.connection_status == PubSubClientConnectionStatus.OK:
                self.connection_status = PubSubClientConnectionStatus.LOST
                for _, status_handler in self.status_handlers:
                    try:
                        await status_handler.on_connection_lost()
                    except BaseException as e:
                        background_exceptions.append(e)

            self.connection_status = PubSubClientConnectionStatus.OK
            for _, status_handler in self.status_handlers:
                try:
                    await status_handler.on_connection_established()
                except BaseException as e:
                    background_exceptions.append(e)

        resp_tracing = b""  # TODO: tracing
        resp_authorization = await self.config.authorize_confirm_missed(
            tracing=resp_tracing,
            topic=topic,
            url=str(request.url),
            now=time.time(),
        )
        resp_authorization_bytes = (
            b"" if resp_authorization is None else resp_authorization.encode("utf-8")
        )
        resp_body = io.BytesIO()
        resp_body.write(
            int(
                SubscriberToBroadcasterStatelessMessageType.RESPONSE_CONFIRM_MISSED
            ).to_bytes(2, "big")
        )
        resp_body.write(len(resp_authorization_bytes).to_bytes(2, "big"))
        resp_body.write(resp_authorization_bytes)
        resp_body.write(len(resp_tracing).to_bytes(2, "big"))
        resp_body.write(resp_tracing)
        return Response(
            status_code=200,
            headers={"Content-Type": "application/octet-stream"},
            content=resp_body.getvalue(),
            background=(
                BackgroundTask(
                    _raise_excs,
                    "failed to inform status handlers about recovery",
                    background_exceptions,
                )
                if background_exceptions
                else None
            ),
        )

    async def register_on_message(
        self, /, *, receiver: PubSubDirectOnMessageWithCleanupReceiver
    ) -> int:
        new_id = 1 if not self.handlers else self.handlers[-1][0] + 1
        self.handlers.append((new_id, receiver))
        return new_id

    async def unregister_on_message(self, /, *, registration_id: int) -> None:
        # seems more likely a more recent handler is being removed, hence search
        # from tail
        idx = len(self.handlers) - 1
        while idx >= 0:
            if self.handlers[idx][0] == registration_id:
                self.handlers.pop(idx)
                return
            idx -= 1

    async def register_status_handler(
        self, /, *, receiver: PubSubDirectConnectionStatusReceiver
    ) -> int:
        self._status_counter += 1
        status_handler_id = self._status_counter
        async with self._status_handlers_lock:
            self.status_handlers.append((status_handler_id, receiver))
            if self.connection_status == PubSubClientConnectionStatus.OK:
                try:
                    await receiver.on_connection_established()
                except BaseException:
                    for idx, (reg_id, _) in enumerate(self.status_handlers):
                        if reg_id == status_handler_id:
                            self.status_handlers.pop(idx)
                    raise
            return status_handler_id

    async def unregister_status_handler(self, /, *, registration_id: int) -> None:
        async with self._status_handlers_lock:
            assert registration_id <= self._status_counter, "invalid registration id"

            # seems more likely a more recent handler is being removed, hence search
            # from tail
            idx = len(self.status_handlers) - 1
            while idx >= 0:
                if self.status_handlers[idx][0] == registration_id:
                    self.status_handlers.pop(idx)
                    return
                idx -= 1
            raise ValueError("invalid registration id")


async def _raise_excs(msg: str, excs: List[BaseException]) -> None:
    raise combine_multiple_exceptions(msg, excs)


if TYPE_CHECKING:
    ___: Type[PubSubClientReceiver] = HttpPubSubClientReceiver


def HttpPubSubClient(config: HttpPubSubConfig) -> PubSubClient:
    """A constructor-like function that creates a pub sub client that connects
    via outgoing HTTP calls and receives notifications via incoming HTTP calls
    """

    async def setup() -> None:
        await config.setup_to_subscriber_auth()
        try:
            await config.setup_to_broadcaster_auth()
        except BaseException:
            await config.teardown_to_subscriber_auth()
            raise

    async def teardown() -> None:
        try:
            await config.teardown_to_broadcaster_auth()
        finally:
            await config.teardown_to_subscriber_auth()

    return PubSubClient(
        HttpPubSubClientConnector(config),
        HttpPubSubClientReceiver(config),
        setup=setup,
        teardown=teardown,
    )



from lonelypsp.auth.config import AuthResult

from lonelypsc.client import PubSubError
from lonelypsc.ws.check_result import CheckResult
from lonelypsc.ws.state import (
    ReceivedMessageType,
    ReceivedMissedMessage,
    ReceivingState,
    StateOpen,
)


def check_receiving_authorizing_missed(state: StateOpen) -> CheckResult:
    """
    Tries to move from receiving state `AUTHORIZING_MISSED` to None

    Raises an error if the authorization result is not "ok"
    """
    if (
        state.receiving is None
        or state.receiving.type != ReceivingState.AUTHORIZING_MISSED
    ):
        return CheckResult.CONTINUE

    if not state.receiving.authorization_task.done():
        return CheckResult.CONTINUE

    result = state.receiving.authorization_task.result()
    if result != AuthResult.OK:
        raise PubSubError(f"authorization failed: {result}")

    state.received.put_nowait(
        ReceivedMissedMessage(
            type=ReceivedMessageType.MISSED, topic=state.receiving.message.topic
        )
    )
    state.receiving = None
    return CheckResult.RESTART



from lonelypsc.ws.check_result import CheckResult
from lonelypsc.ws.state import ReceivingState, StateOpen


def check_receiving_decompressing(state: StateOpen) -> CheckResult:
    """
    Tries to move from receiving notify state `DECOMPRESSING` to None

    Raises an exception if the decompression task fails
    """
    if state.receiving is None or state.receiving.type != ReceivingState.DECOMPRESSING:
        return CheckResult.CONTINUE

    if not state.receiving.task.done():
        return CheckResult.CONTINUE

    msg = state.receiving.task.result()
    state.received.put_nowait(msg)
    state.receiving = None
    return CheckResult.RESTART



from types import TracebackType
from typing import Optional, Type

from lonelypsc.ws.compressor import CompressorReady
from lonelypsc.ws.state import StateOpen

try:
    import zstandard
except ImportError:
    ...


class CompressorReservation:
    """A convenience object for generating or reserving an actual object
    capable of compressing data from a ready compressor configuration

    Acts as a synchronous context manager; must be entered to use the
    compressor and exited to either dispose of the compressor or return it
    to the pool based on the current pool size

    Doesn't return the compressor object to the pool if an exception is passed
    to exit, in case the compressor is in an invalid state
    """

    def __init__(self, compressor: CompressorReady, *, max_compressors: int) -> None:
        self.compressor_configuration = compressor
        """the configuration for the compressor that we generate"""

        self.max_compressors = max_compressors
        """the max number of compressors to hold in the pool"""

        self._compressor: "Optional[zstandard.ZstdCompressor]" = None
        """the compressor object that we are holding onto"""

    def __enter__(self) -> "zstandard.ZstdCompressor":
        assert self._compressor is None, "not re-entrant"

        if self.compressor_configuration.compressors:
            self._compressor = self.compressor_configuration.compressors.pop()
        else:
            self._compressor = zstandard.ZstdCompressor(
                level=self.compressor_configuration.level,
                dict_data=self.compressor_configuration.data,
                write_checksum=False,
                write_content_size=False,
                write_dict_id=False,
            )

        return self._compressor

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        if self._compressor is None or exc_type is not None:
            return

        if len(self.compressor_configuration.compressors) < self.max_compressors:
            self.compressor_configuration.compressors.append(self._compressor)

        self._compressor = None


class DecompressorReservation:
    """A convenience object for generating or reserving an actual object
    capable of decompressing data from a ready compressor configuration

    Doesn't return the decompressor object to the pool if an exception is passed
    to exit, in case the decompressor is in an invalid state
    """

    def __init__(
        self,
        compressor: CompressorReady,
        *,
        max_window_size: int,
        max_decompressors: int,
    ) -> None:
        self.compressor_configuration = compressor
        """the configuration for the compressor that we generate"""

        self.max_decompressors = max_decompressors
        """the max number of decompressors to hold in the pool"""

        self.max_window_size = max_window_size
        """the max window size if we need to create a new decompressor"""

        self._decompressor: "Optional[zstandard.ZstdDecompressor]" = None
        """the decompressor object that we are holding onto"""

    def __enter__(self) -> "zstandard.ZstdDecompressor":
        assert self._decompressor is None, "not re-entrant"

        if self.compressor_configuration.decompressors:
            self._decompressor = self.compressor_configuration.decompressors.pop()
        else:
            self._decompressor = zstandard.ZstdDecompressor(
                dict_data=self.compressor_configuration.data,
                max_window_size=self.max_window_size,
            )

        return self._decompressor

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        if self._decompressor is None or exc_type is not None:
            return

        if len(self.compressor_configuration.decompressors) < self.max_decompressors:
            self.compressor_configuration.decompressors.append(self._decompressor)

        self._decompressor = None


def reserve_compressor(
    state: StateOpen, compressor: CompressorReady
) -> CompressorReservation:
    """Recommended way to initialize a compressor reservation, which will handle
    extracting any required configuration options
    """
    return CompressorReservation(compressor, max_compressors=5)


def reserve_decompressor(
    state: StateOpen, compressor: CompressorReady
) -> DecompressorReservation:
    """Recommended way to initialize a decompressor reservation, which will handle
    extracting any required configuration options
    """
    return DecompressorReservation(
        compressor,
        max_window_size=state.config.decompression_max_window_size,
        max_decompressors=5,
    )



import asyncio

from lonelypsp.stateful.constants import SubscriberToBroadcasterStatefulMessageType

from lonelypsc.ws.check_result import CheckResult
from lonelypsc.ws.handlers.open.sends.confirm_receive import send_confirm_receive
from lonelypsc.ws.handlers.open.sends.continue_receive import send_continue_receive
from lonelypsc.ws.state import SendingSimple, SendingState, StateOpen


def check_unsent_acks(state: StateOpen) -> CheckResult:
    """Checks if the left item from unsent acks can be moved to send_task;
    doing so if possible and nothing otherwise
    """
    if state.sending is not None:
        return CheckResult.CONTINUE

    if not state.unsent_acks:
        return CheckResult.CONTINUE

    next_ack = state.unsent_acks.popleft()
    ack_co = (
        send_confirm_receive(state, next_ack)
        if next_ack.type == SubscriberToBroadcasterStatefulMessageType.CONFIRM_RECEIVE
        else send_continue_receive(state, next_ack)
    )

    state.sending = SendingSimple(
        type=SendingState.SIMPLE,
        task=asyncio.create_task(ack_co),
    )
    return CheckResult.RESTART



import asyncio
from typing import TYPE_CHECKING

from lonelypsp.stateful.messages.confirm_notify import B2S_ConfirmNotify

from lonelypsc.client import PubSubError, PubSubIrrecoverableError
from lonelypsc.ws.handlers.open.messages.protocol import MessageChecker
from lonelypsc.ws.internal_callbacks import finalize_internal_callback
from lonelypsc.ws.state import (
    InternalMessageStateAcknowledged,
    InternalMessageStateType,
    SendingState,
    StateOpen,
)


def check_confirm_notify(state: StateOpen, message: B2S_ConfirmNotify) -> None:
    """Handles the subscriber receiving the a message from the broadcaster that
    it has successfully processed a message the subscriber previously sent with
    NOTIFY or NOTIFY_STREAM, raising an error if this isn't the acknowledgement
    the subscriber was expecting.
    """
    try:
        expected = state.expected_acks.popleft()
    except IndexError:
        raise PubSubError(f"{message.type} received when not expecting any acks")

    if expected.type != message.type:
        raise PubSubError(f"expected {expected.type}, got {message.type}")
    if expected.identifier != message.identifier:
        raise PubSubError(
            f"expected {message.type} {expected.identifier!r}, got {message.identifier!r}"
        )

    try:
        internal_message = state.sent_notifications.popleft()
    except IndexError:
        raise PubSubIrrecoverableError(
            f"invariant violated: expected {expected.type}, but no corresponding internal message"
        )

    if internal_message.identifier != message.identifier:
        raise PubSubIrrecoverableError(
            f"invariant violated: expected {expected.type} {message.identifier!r}, got {internal_message.identifier!r}"
        )

    if (
        state.sending is not None
        and state.sending.type == SendingState.INTERNAL_MESSAGE
        and state.sending.internal_message.identifier == message.identifier
    ):
        assert (
            state.sending.internal_message is internal_message
        ), "invariant violated: multiple copies with same identifier"
        state.backgrounded.add(state.sending.task)
        state.sending = None

    state.backgrounded.add(
        asyncio.create_task(
            finalize_internal_callback(
                internal_message.callback,
                InternalMessageStateAcknowledged(
                    type=InternalMessageStateType.ACKNOWLEDGED,
                    notified=message.subscribers,
                ),
            )
        )
    )


if TYPE_CHECKING:
    _: MessageChecker[B2S_ConfirmNotify] = check_confirm_notify



from lonelypsc.ws.check_result import CheckResult
from lonelypsc.ws.internal_callbacks import sweep_internal_message
from lonelypsc.ws.state import StateOpen


def check_sent_notifications(state: StateOpen) -> CheckResult:
    """Sweeps all the sent notifications (that haven't been acknowledged), raising
    an error if any of the backgrounded internal message state callback tasks
    failed.
    """
    for msg in state.sent_notifications:
        sweep_internal_message(msg)
    return CheckResult.CONTINUE



from typing import Protocol

from lonelypsc.ws.state import State


class StateHandler(Protocol):
    """Describes something capable of making some progress within the
    websocket state machine, returning the next state (or the same state),
    such that if we call the appropriate state handlers in a loop we eventually
    end in StateClosed

    This protocol is mainly to ensure we are being internally consistent with
    the function signature

    The signature is a generic State rather than a contravariant typevar because
    in order for this to be useful we need to collect the state handlers in a dict,
    and afaik theres no way to type a dictionary whose keys and values are related,
    e.g., `Dict[K => StateHandler[K]]` (this is doable in e.g typescript)
    """

    async def __call__(self, state: State) -> State: ...



from lonelypsc.ws.state import StateOpen


def make_for_send_websocket_url_and_change_counter(state: StateOpen) -> str:
    """Creates the websocket URL that the subscriber should use when contacting
    the broadcaster. The result value is always unique and deterministic, and
    should be independently calculated by the broadcaster (and checked for a
    match)
    """
    ctr = state.subscriber_counter
    state.subscriber_counter -= 1
    return f"websocket:{state.nonce_b64}:{ctr:x}"


def make_for_receive_websocket_url_and_change_counter(state: StateOpen) -> str:
    """Creates the websocket URL that the broadcaster should use when contacting
    the subscriber. The result value is always unique and deterministic, and
    should be independently calculated by the subscriber (and checked for a
    match)
    """
    ctr = state.broadcaster_counter
    state.broadcaster_counter += 1
    return f"websocket:{state.nonce_b64}:{ctr:x}"



import asyncio

from lonelypsc.ws.check_result import CheckResult
from lonelypsc.ws.handlers.open.send_internal_message import send_internal_message
from lonelypsc.ws.internal_callbacks import sweep_internal_message
from lonelypsc.ws.state import SendingInternalMessage, SendingState, StateOpen


def check_resending_notifications(state: StateOpen) -> CheckResult:
    """First, sweeps all the resending notifications, raising an error
    if any of the backgrounded internal message state callback tasks
    failed.

    Then, checks if the left item from resending_notifications can be moved
    to sending, doing so if possible and nothing otherwise
    """
    if not state.resending_notifications:
        return CheckResult.CONTINUE

    for msg in state.resending_notifications:
        sweep_internal_message(msg)

    if state.sending is not None:
        return CheckResult.CONTINUE

    msg = state.resending_notifications.pop()
    state.sending = SendingInternalMessage(
        type=SendingState.INTERNAL_MESSAGE,
        task=asyncio.create_task(send_internal_message(state, msg)),
        internal_message=msg,
    )
    return CheckResult.RESTART



from typing import Literal, TypedDict, Union


class WSMessageBytes(TypedDict):
    type: Literal["websocket.receive"]
    bytes: bytes


class WSMessageText(TypedDict):
    type: Literal["websocket.receive"]
    text: str


class WSMessageDisconnect(TypedDict):
    type: Literal["websocket.disconnect"]
    code: int
    reason: str


WSMessage = Union[WSMessageBytes, WSMessageText, WSMessageDisconnect]
"""A better type for Message from websockets"""



from typing import TYPE_CHECKING, Dict

from lonelypsc.client import PubSubCancelRequested
from lonelypsc.ws.handlers.closing import handle_closing
from lonelypsc.ws.handlers.configuring import handle_configuring
from lonelypsc.ws.handlers.connecting import handle_connecting
from lonelypsc.ws.handlers.open.handler import handle_open
from lonelypsc.ws.handlers.protocol import StateHandler
from lonelypsc.ws.handlers.waiting_retry import handle_waiting_retry
from lonelypsc.ws.state import State, StateClosed, StateType

_handlers: Dict[StateType, StateHandler] = {
    StateType.CONNECTING: handle_connecting,
    StateType.CONFIGURING: handle_configuring,
    StateType.WAITING_RETRY: handle_waiting_retry,
    StateType.OPEN: handle_open,
    StateType.CLOSING: handle_closing,
}


async def handle_any(state: State) -> State:
    """Handle any state by delegating to the appropriate handler. This will raise
    a KeyError for the CLOSED state, which does not need any further handling.

    Raises KeyError if no handler is found for the state type.
    """
    try:
        return await _handlers[state.type](state)
    except PubSubCancelRequested:
        return StateClosed(type=StateType.CLOSED)


if TYPE_CHECKING:
    _: StateHandler = handle_any



import asyncio

import aiohttp

from lonelypsc.config.config import PubSubBroadcasterConfig
from lonelypsc.config.ws_config import WebsocketPubSubConfig


def make_websocket_connect_task(
    config: WebsocketPubSubConfig,
    broadcaster: PubSubBroadcasterConfig,
    client_session: aiohttp.ClientSession,
) -> asyncio.Task[aiohttp.ClientWebSocketResponse]:
    """Creates the standard task to connect to the given broadcaster within the
    given session, usually for creating a CONNECTING state

    Args:
        config (WebsocketPubSubConfig): how the subscriber is configured
        broadcaster (PubSubBroadcasterConfig): the broadcaster to connect to
        client_session (aiohttp.ClientSession): the session to use for the connection
    """
    return asyncio.create_task(
        client_session.ws_connect(
            broadcaster["host"] + "/v1/websocket",
            # WARN: do not use ClientWSTimeout ws_receive, which will ignore
            # heartbeats, meaning it will timeout unless there are actual
            # notify/subscribe messages being sent. the heartbeat interval
            # is acting as our receive timeout
            timeout=aiohttp.ClientWSTimeout(
                ws_receive=None, ws_close=config.websocket_close_timeout
            ),
            heartbeat=config.websocket_heartbeat_interval,
        )
    )
